import numpy as np
import pandas as pd
import pathlib
from typing import overload

from glassesTools import annotation, naming as gt_naming, timestamps, video_utils

from . import episode, naming

def get_cols(do_time_stretch: bool):
    cols    = ['t_ref','t_this','offset']
    if do_time_stretch:
        cols += ['t_ref_elapsed','diff_offset','stretch_fac']
    else:
        cols += ['mean_off']

def get_sync_for_recs(working_dir: str|pathlib.Path, recs: str|list[str], ref_rec: str, do_time_stretch: bool, average_recordings: list[str], missing_ref_coding_ok=False):
    working_dir  = pathlib.Path(working_dir)
    if isinstance(recs,str):
        recs = [recs]
    ref_episodes = get_coding_file(working_dir / ref_rec, missing_ref_coding_ok)
    if ref_episodes is None:
        return None
    video_ts_ref = timestamps.VideoTimestamps(working_dir / ref_rec / gt_naming.frame_timestamps_fname)

    if do_time_stretch:
        if len(ref_episodes)<2:
            if missing_ref_coding_ok:
                return None
            raise ValueError(f"You requested to do time stretching when syncing the recordings, but there is only one camera sync point in the reference recording. At least two sync points are required for time stretching")
        for r in average_recordings:
            if r not in recs:
                raise ValueError(f'Recording {r} not found')
            if r==ref_rec:
                raise ValueError(f'Recording {r} is the reference recording for sync, should not be specified in study_config.sync_average_recordings')

    index = pd.MultiIndex.from_product([recs, range(len(ref_episodes))], names=['recording','interval'])
    sync  = pd.DataFrame(columns=get_cols(do_time_stretch), dtype=float, index=index)

    # collect timestamps for the recordings
    for r in recs:
        # get interval coding for this recording
        episodes = get_coding_file(working_dir / r, missing_ref_coding_ok)
        if episodes is None and missing_ref_coding_ok:
            return None

        # check intervals
        if len(episodes)!=len(ref_episodes):
            if missing_ref_coding_ok:
                return None
            raise ValueError(f"The number of sync points for this recording ({len(episodes)}, {r}) is not equal to that for the reference recording ({len(ref_episodes)}, {ref_rec}). Cannot continue, fix your coding")

        # get time information
        video_ts = timestamps.VideoTimestamps(working_dir / r / gt_naming.frame_timestamps_fname)

        # get timestamps corresponding to sync frames
        for ival in range(len(episodes)):
            sync.loc[(r,ival),'t_ref']  = video_ts_ref.get_timestamp(ref_episodes[ival])/1000.  # ms -> s
            sync.loc[(r,ival),'t_this'] =   video_ts  .get_timestamp(    episodes[ival])/1000.  # ms -> s
            sync.loc[(r,ival),'offset'] = sync.loc[(r,ival),'t_ref']-sync.loc[(r,ival),'t_this']
        if not do_time_stretch:
            # no time stretching, get average offset. Applies to whole file, store only for first interval
            sync.loc[(r,0),'mean_off'] = sync.loc[(r,slice(None)),'offset'].mean()

    if do_time_stretch:
        # get stretch factor for each interval between two sync points
        if average_recordings:
            recs_gr = [r for r in recs if r not in average_recordings]
            recs_gr.append(list(average_recordings))
        else:
            recs_gr = recs
        for r in recs_gr:
            for ival in range(len(ref_episodes)-1):
                t_ref = sync.loc[(r,ival+1),'t_ref' ]
                offset= sync.loc[(r,ival+1),'offset']
                if not isinstance(t_ref,float) and 'interval' in t_ref.index.names:
                    t_ref = t_ref .droplevel('interval')
                    offset= offset.droplevel('interval')
                sync.loc[(r,ival),'t_ref_elapsed'] = t_ref-sync.loc[(r,ival),'t_ref']
                sync.loc[(r,ival),'diff_offset']   = offset-sync.loc[(r,ival),'offset']
                sync.loc[(r,ival),'stretch_fac']   = sync.loc[(r,ival),'diff_offset'].mean()/sync.loc[(r,ival),'t_ref_elapsed'].mean()
    return sync

def apply_sync(rec: str,
               sync: pd.DataFrame,
               data_timestamps: np.ndarray|None,
               reference_video_timestamps: np.ndarray,
               do_time_stretch,
               stretch_which: str):
    reference_video_timestamps  = np.array(reference_video_timestamps).copy()
    new_reference_video_timestamps = reference_video_timestamps.copy()
    rt_start, rt_end            = reference_video_timestamps.min(), reference_video_timestamps.max()
    if has_data_ts := data_timestamps is not None:
        data_timestamps     = np.array(data_timestamps).copy()
        dt_start, dt_end    = data_timestamps.min(), data_timestamps.max()
        new_data_timestamps = data_timestamps.copy()
    else:
        dt_start, dt_end    = None, None
    num_reference_episodes      = sync.loc[rec].shape[0]
    if do_time_stretch:
        for ival in range(num_reference_episodes-1):
            # set up the problem - piecewise linear scale
            # 1. get known good location for this interval, and the stretch factor
            pivot       = (sync.loc[(rec,ival),'t_ref'] if stretch_which=='ref' else sync.loc[(rec,ival),'t_this'])*1000.   # s -> ms
            stretch_fac = sync.loc[(rec,ival),'stretch_fac']
            # 2. determine data range to apply stretch for this interval to
            if ival==0:
                # first interval, apply all the way from start of data
                d_start = dt_start
                r_start = rt_start
            else:
                d_start = r_start = sync.loc[(rec,ival),'t_ref']
            if ival==num_reference_episodes-2:  # NB: zero-based indexing
                # last interval, apply all the way to end of data
                d_end = dt_end
                r_end = rt_end
            else:
                d_end = r_end = sync.loc[(rec,ival+1),'t_ref']
            # calculate new timestamps
            if stretch_which=='ref':
                # 1. first translate gaze ts to reference timestamps
                if has_data_ts:
                    data_sel = (data_timestamps >= d_start) & (data_timestamps <= d_end)
                    new_data_timestamps[data_sel] += sync.loc[(rec,ival),'offset']*1000.   # s -> ms
                # 2. apply scaling
                data_sel = (reference_video_timestamps >= r_start) & (reference_video_timestamps <= r_end)
                new_reference_video_timestamps[data_sel] = (reference_video_timestamps[data_sel]-pivot)*(1-stretch_fac)+pivot
            elif stretch_which=='other':
                if has_data_ts:
                    data_sel = (data_timestamps >= d_start) & (data_timestamps <= d_end)
                    new_data_timestamps[data_sel] = (data_timestamps[data_sel]-pivot)*(1+stretch_fac)+pivot
                    new_data_timestamps[data_sel] += sync.loc[(rec,ival),'offset']*1000.   # s -> ms
                # else nothing to do...
    else:
        if has_data_ts:
            new_data_timestamps += sync.loc[(rec,0),'mean_off']*1000.   # s -> ms
    if has_data_ts:
        fr_ref = video_utils.timestamps_to_frame_number(new_data_timestamps,new_reference_video_timestamps,trim=True)['frame_idx'].to_numpy()
    else:
        fr_ref = None
    return new_data_timestamps, new_reference_video_timestamps, fr_ref

def get_coding_file(working_dir: str|pathlib.Path, missing_ref_coding_ok=False):
    working_dir  = pathlib.Path(working_dir)
    coding_file = working_dir / naming.coding_file
    if not coding_file.is_file():
        if missing_ref_coding_ok:
            return None
        raise FileNotFoundError(f'A coding file must be available for the recording ({working_dir.name}) to run sync_to_ref, but it is not. Run code_episodes and code at least one {annotation.tooltip_map[annotation.Event.Sync_Camera]}. Not found: {coding_file}')
    episodes = episode.list_to_marker_dict(episode.read_list_from_file(coding_file))[annotation.Event.Sync_Camera]
    episodes = [x[0] for x in episodes] # remove inner wrapping list, there are only single values in it anyway
    if not episodes:
        if missing_ref_coding_ok:
            return None
        raise ValueError(f'No {annotation.tooltip_map[annotation.Event.Sync_Camera]}s found for this recording ({working_dir.name}). Run code_episodes and code at least one {annotation.tooltip_map[annotation.Event.Sync_Camera]}.')
    return episodes

def get_episode_frame_indices_from_ref(working_dir: str|pathlib.Path, event: annotation.Event, rec: str, ref_rec:str, all_recs: list[str], do_time_stretch: bool, average_recordings: list[str], stretch_which: str, extra_fr=0, missing_ref_coding_ok=False) -> list[list[int]]:
    working_dir  = pathlib.Path(working_dir)
    ref_coding_file = working_dir.parent / ref_rec / naming.coding_file
    if not ref_coding_file.is_file():
        if missing_ref_coding_ok:
            return [[]]
        raise FileNotFoundError(f'The coding file for the reference recording is not found, cannot continue ("{ref_coding_file}").')
    ref_episodes = episode.list_to_marker_dict(episode.read_list_from_file(ref_coding_file))
    if event not in ref_episodes:
        if missing_ref_coding_ok:
            return [[]]
        raise KeyError(f'Trying to get {annotation.tooltip_map[event]}s from the reference recording ({ref_rec}), but the coding file for this reference recording doesn\'t contain any {annotation.tooltip_map[event]}s')
    # get sync and timestamp info we need to transform reference frames indices to frame indices of this recording
    sync = get_sync_for_recs(working_dir.parent, all_recs, ref_rec, do_time_stretch, average_recordings, missing_ref_coding_ok)
    if sync is None:
        return [[]]
    video_ts_ref = timestamps.VideoTimestamps(working_dir.parent / ref_rec / gt_naming.frame_timestamps_fname)
    video_ts     = timestamps.VideoTimestamps(working_dir / gt_naming.frame_timestamps_fname)
    # get frame indices in this recording's video corresponding to each of the reference frames
    frame_idx = reference_frames_to_video(rec, sync, ref_episodes[event], video_ts.timestamps, video_ts_ref.timestamps, do_time_stretch, stretch_which)
    return [[i+e for i,e in zip(ifs, [-extra_fr, extra_fr])] for ifs in frame_idx]   # expand by extra_fr frames on each edge

@overload
def reference_frames_to_video(rec: str, sync: pd.DataFrame, fr_idxs: list[int], video_ts: list[float]|np.ndarray, video_ts_ref: list[float]|np.ndarray, do_time_stretch: bool, stretch_which: str) -> list[int]: ...
@overload
def reference_frames_to_video(rec: str, sync: pd.DataFrame, fr_idxs: list[list[int]], video_ts: list[float]|np.ndarray, video_ts_ref: list[float]|np.ndarray, do_time_stretch: bool, stretch_which: str) -> list[list[int]]: ...
def reference_frames_to_video(rec: str, sync: pd.DataFrame, fr_idxs: list[int]|list[list[int]], this_video_ts: list[float]|np.ndarray, video_ts_ref: list[float]|np.ndarray, do_time_stretch: bool, stretch_which: str) -> list[int]|list[list[int]]:
    if not fr_idxs:
        return []

    # get the video's timestamps in time of the reference video
    this_video_ts_ref, video_ts_ref, _ = apply_sync(rec, sync, this_video_ts, video_ts_ref, do_time_stretch, stretch_which)

    # get where (which frame) each of this video's timestamps occur in the reference video, given the sync info
    # (fr_idx_ref contains the reference frame_idxs corresponding to this video's frames, video_ts)
    fr_idx_ref = video_utils.timestamps_to_frame_number(video_ts_ref, this_video_ts_ref, trim=True)['frame_idx'].to_numpy()
    fr_idx_ref[video_ts_ref<this_video_ts_ref[0]] = -1
    # in case only the first fr_idx is trimmed, a little bit of leeway is ok
    if fr_idx_ref.size>=2 and fr_idx_ref[0]==-1 and fr_idx_ref[1]>0:
        ifi = np.mean(np.diff(video_ts_ref))
        # that means, do assign a frame to the first frame if its a usual frame (judged by ifi)
        if video_ts_ref[1]-video_ts_ref[0] < ifi*1.2:
            fr_idx_ref[0] = fr_idx_ref[1]-1

    return fr_idx_ref[fr_idxs].tolist()

@overload
def video_frames_to_reference(rec: str, sync: pd.DataFrame, fr_idxs: list[int], video_ts: list[float]|np.ndarray, video_ts_ref: list[float]|np.ndarray, do_time_stretch: bool, stretch_which: str) -> list[int]: ...
@overload
def video_frames_to_reference(rec: str, sync: pd.DataFrame, fr_idxs: list[list[int]], video_ts: list[float]|np.ndarray, video_ts_ref: list[float]|np.ndarray, do_time_stretch: bool, stretch_which: str) -> list[list[int]]: ...
def video_frames_to_reference(rec: str, sync: pd.DataFrame, fr_idxs: list[int]|list[list[int]], this_video_ts: list[float]|np.ndarray, video_ts_ref: list[float]|np.ndarray, do_time_stretch: bool, stretch_which: str) -> list[int]|list[list[int]]:
    if not fr_idxs:
        return []

    # get the video's timestamps in time of the reference video
    this_video_ts_ref, video_ts_ref, _ = apply_sync(rec, sync, this_video_ts, video_ts_ref, do_time_stretch, stretch_which)

    # get where (which frame) each of the reference video frames occur in this video, given the sync info
    # (fr_idx contains this video's frame_idxs corresponding to this reference's frames, video_ts)
    fr_idx = video_utils.timestamps_to_frame_number(this_video_ts_ref, video_ts_ref, trim=True)['frame_idx'].to_numpy()
    fr_idx[this_video_ts_ref<video_ts_ref[0]] = -1

    return fr_idx[fr_idxs].tolist()

def smooth_video_frames_indices(fr_idxs: list[int]):
    # detect plateaus of N samples followed by a step of N samples
    # that may occur if sampling rates of two videos are unmatched when syncing them.
    # while the plateaus are correct for "nearest" frame logic, for display purposes
    # a smoothed version of the higher framerate video may be wanted
    fr_idxs = np.array(fr_idxs)

    # get steps in frame_signal
    d       = np.diff(fr_idxs)
    if np.any(d<0):
        # below logic assumes a increasing frame index array
        return fr_idxs.tolist()

    # find where the plateaus are (consecutive steps of 0)
    vals    = np.pad((d==0).astype(int), (1, 1), 'constant', constant_values=(0, 0))
    d2      = np.diff(vals)
    starts  = np.nonzero(d2 == 1)[0]
    ends    = np.nonzero(d2 == -1)[0]
    # ensure we're not out of bounds
    if ends.size>0 and ends[-1]==len(d):
        starts = starts[:-1]
        ends   = ends[:-1]

    # for each plateau, see how long it is, and what the step to the next value after it is
    plateau_len = ends-starts+1
    step_at_end = d[ends]

    # select those plateaus whose step after is the same size as the plateau length (e.g. frame indices [235 236 236 238], a plateau of length 2, and a step of two frames thereafter)
    # those we can fix up
    to_fix = np.nonzero(np.logical_and(plateau_len==step_at_end, starts!=-1))[0]

    # fix em
    for i in to_fix:
        fr_start = fr_idxs[starts[i]]
        fr_end   = fr_idxs[ends[i]+1]
        new_vals = list(range(fr_start,fr_end))
        fr_idxs[starts[i]:ends[i]+1] = new_vals

    return fr_idxs.tolist()


import pathlib

from glassesTools import annotation, gaze_headref, gaze_worldref, naming as gt_naming, ocv, pose as gt_pose, process_pool, propagating_thread
from glassesTools.gui import worldgaze as worldgaze_gui
from glassesTools.gui.video_player import GUI

from .. import config, episode, naming, plane, process, session, synchronization


def run(working_dir: str|pathlib.Path, config_dir: str|pathlib.Path = None, show_visualization=False, show_planes=True, show_only_intervals=True, progress_indicator: process_pool.JobProgress=None, **study_settings):
    # if show_visualization, each frame is shown in a viewer, overlaid with info about detected planes and projected gaze
    # if show_poster, gaze in space of each plane is also drawn in a separate windows
    # if show_only_intervals, only the coded mapping episodes (if available) are shown in the viewer while the rest of the scene video is skipped past
    working_dir = pathlib.Path(working_dir)
    if config_dir is None:
        config_dir = config.guess_config_dir(working_dir)
    config_dir  = pathlib.Path(config_dir)

    print(f'processing: {working_dir.parent.name}/{working_dir.name}')

    # if we need gui, we run processing in a separate thread (GUI needs to be on the main thread for OSX, see https://github.com/pthom/hello_imgui/issues/33)
    if show_visualization:
        gui = GUI(use_thread = False)
        gui.add_window(working_dir.name)
        gui.set_show_controls(True)
        gui.set_show_play_percentage(True)
        gui.set_show_action_tooltip(True)

        proc_thread = propagating_thread.PropagatingThread(target=do_the_work, args=(working_dir, config_dir, gui, show_planes, show_only_intervals, progress_indicator), kwargs=study_settings, cleanup_fun=gui.stop)
        proc_thread.start()
        gui.start()
        proc_thread.join()
    else:
        do_the_work(working_dir, config_dir, None, False, False, progress_indicator, **study_settings)


def do_the_work(working_dir: pathlib.Path, config_dir: pathlib.Path, gui: GUI, show_planes: bool, show_only_intervals: bool, progress_indicator: process_pool.JobProgress, **study_settings):
    # progress indicator
    if progress_indicator is None:
        progress_indicator = process_pool.JobProgress(printer=lambda x: print(x))
    progress_indicator.set_unit('samples')
    progress_indicator.set_start_time_to_now()

    # get settings for the study
    study_config = config.read_study_config_with_overrides(config_dir, {config.OverrideLevel.Session: working_dir.parent, config.OverrideLevel.Recording: working_dir}, **study_settings)

    # get info about recording
    rec_def = study_config.session_def.get_recording_def(working_dir.name)
    if rec_def.type!=session.RecordingType.Eye_Tracker:
        raise ValueError(f'You can only run gaze_to_plane on eye tracker recordings, not on a {str(rec_def.type).split(".")[1]} recording')

    # get episodes for which to transform gaze
    episodes = episode.list_to_marker_dict(episode.read_list_from_file(working_dir / naming.coding_file), study_config.episodes_to_code)
    # trial episodes are gotten from the reference recording if there is one and this is not the reference recording
    if study_config.sync_ref_recording and rec_def.name!=study_config.sync_ref_recording:
        if annotation.Event.Trial in episodes and episodes[annotation.Event.Trial]:
            raise ValueError(f'Trial episodes are gotten from the reference recording ({study_config.sync_ref_recording}) and should not be coded for this recording ({rec_def.name})')
        if annotation.Event.Trial in study_config.episodes_to_code:
            all_recs = [r.name for r in study_config.session_def.recordings]
            episodes[annotation.Event.Trial] = synchronization.get_episode_frame_indices_from_ref(working_dir, annotation.Event.Trial, rec_def.name, study_config.sync_ref_recording, all_recs, study_config.sync_ref_do_time_stretch, study_config.sync_ref_average_recordings, study_config.sync_ref_stretch_which)

    # we transform to map to plane for validate and trial episodes, set it up
    episodes_to_proc = [annotation.Event.Validate, annotation.Event.Trial]
    if annotation.Event.Sync_ET_Data in study_config.episodes_to_code and study_config.get_cam_movement_for_et_sync_method=='plane':
        episodes_to_proc.append(annotation.Event.Sync_ET_Data)
    mapping_setup: dict[str, list[list[int]]] = {}
    for e in episodes_to_proc:
        if e in study_config.planes_per_episode:
            for p in study_config.planes_per_episode[e]:
                if p not in mapping_setup:
                    mapping_setup[p] = []
                mapping_setup[p].extend(episodes[e])
    mapping_setup = {p:sorted(mapping_setup[p], key = lambda x: x[0]) for p in mapping_setup}

    planes: dict[str,plane.Plane] = {}
    for p in mapping_setup:
        p_def = [pl for pl in study_config.planes if pl.name==p][0]
        planes[p] = plane.get_plane_from_definition(p_def, config_dir/p)

    # load gaze data and poses
    processing_intervals = [e for p in mapping_setup for e in mapping_setup[p]] # NB: doesn't need to be sorted
    should_load_part = not gui or show_only_intervals
    head_gazes = gaze_headref.read_dict_from_file(working_dir / gt_naming.gaze_data_fname, processing_intervals if should_load_part else None, ts_column_suffixes=['VOR', ''])[0]
    poses = {p:gt_pose.read_dict_from_file(working_dir/f'{naming.plane_pose_prefix}{p}.tsv', mapping_setup[p] if should_load_part else None) for p in mapping_setup}

    # prep progress indicator
    total = sum(len(head_gazes[f]) for p in poses for f in poses[p] if f in head_gazes)
    progress_indicator.set_total(total)
    progress_indicator.set_intervals(int(total/200), int(total/200))
    # get camera calibration info
    camera_params = ocv.CameraParams.read_from_file(working_dir / gt_naming.scene_camera_calibration_fname)

    # transform gaze to plane(s)
    plane_gazes: dict[str, dict[int,list[gaze_worldref.Gaze]]] = {}
    for p in planes:
        plane_gazes[p] = gaze_worldref.from_head(poses[p], head_gazes, camera_params, progress_indicator.update)
        gaze_worldref.write_dict_to_file(plane_gazes[p], working_dir/f'{naming.world_gaze_prefix}{p}.tsv', skip_missing=True)

    # update state
    session.update_action_states(working_dir, process.Action.GAZE_TO_PLANE, process_pool.State.Completed, study_config)

    # done if no visualization wanted
    if gui is None:
        return

    in_video = session.read_recording_info(working_dir, rec_def.type)[1]
    worldgaze_gui.show_visualization(
        in_video, working_dir / gt_naming.frame_timestamps_fname, working_dir / gt_naming.scene_camera_calibration_fname,
        planes, poses, head_gazes, plane_gazes,
        {e:episodes[e] for e in episodes_to_proc if e in episodes},
        gui, show_planes, show_only_intervals, 8
    )


import pathlib
import numpy as np
import shutil
import copy

from glassesTools import annotation, marker as gt_marker, process_pool

from .. import config, episode, naming, process, session


def run(working_dir: str|pathlib.Path, config_dir: str|pathlib.Path = None, **study_settings):
    working_dir = pathlib.Path(working_dir) # working directory of a session, not of a recording
    if config_dir is None:
        config_dir = config.guess_config_dir(working_dir)
    config_dir  = pathlib.Path(config_dir)
    print(f'processing: {working_dir.name}')

    # get settings for the study
    study_config = config.read_study_config_with_overrides(config_dir, {config.OverrideLevel.Session: working_dir.parent, config.OverrideLevel.Recording: working_dir}, **study_settings)
    if not study_config.auto_code_episodes:
        raise ValueError(f'Nothing to do, no automatic episode coding is defined for this study')
    rec_def = study_config.session_def.get_recording_def(working_dir.name)

    if study_config.sync_ref_recording and rec_def.name!=study_config.sync_ref_recording:
        # Trial events are only coded for the reference recording
        study_config.auto_code_episodes.pop(annotation.Event.Trial, None)
        if not study_config.auto_code_episodes:
            raise RuntimeError(f'Nothing to do, auto_code_episodes is defined only for trials and you have a sync_ref_recording ({study_config.sync_ref_recording}), but this recording ({rec_def.name}) is another one')

    # get already coded interval(s), if any
    coding_file = working_dir / naming.coding_file
    if coding_file.is_file():
        episodes = episode.list_to_marker_dict(episode.read_list_from_file(coding_file), study_config.episodes_to_code)
        # flatten
        for e in episodes:
            episodes[e] = [i for iv in episodes[e] for i in iv]
    else:
        episodes = episode.get_empty_marker_dict(study_config.episodes_to_code)
    episodes_original = copy.deepcopy(episodes)

    # get marker files
    all_marker_ids = [study_config.auto_code_episodes[e]['start_markers']+study_config.auto_code_episodes[e]['end_markers'] for e in study_config.auto_code_episodes]
    all_marker_ids = {m for ms in all_marker_ids for m in ms}
    file_missing   = [not gt_marker.get_file_name(m.m_id, m.aruco_dict_id, working_dir).is_file() for m in all_marker_ids]
    if any(file_missing):
        file_missing = [gt_marker.get_file_name(m.m_id, m.aruco_dict_id, None) for m,miss in zip(all_marker_ids,file_missing) if miss]
        missing_str  = '\n- '.join(file_missing)
        raise FileNotFoundError(f'The following marker files were not found:\n- {missing_str}')
    ori_markers = {m: gt_marker.read_dataframe_from_file(m.m_id, m.aruco_dict_id, working_dir) for m in all_marker_ids}
    # recode so we have a boolean with when markers are present
    ori_markers = {m: gt_marker.code_for_presence(ori_markers[m], allow_failed=True) for m in ori_markers if not ori_markers[m].empty}
    # marker presence signal only contains marker detections (True). We need to fill the gaps in between detections with False (not detected) so we have a continuous signal without gaps
    ori_markers = {m: gt_marker.expand_detection(ori_markers[m], fill_value=False) for m in ori_markers}
    # now auto code indicated intervals
    for e in study_config.auto_code_episodes:
        markers = {m: copy.deepcopy(ori_markers[m]) for m in study_config.auto_code_episodes[e]['start_markers']+study_config.auto_code_episodes[e]['end_markers']}
        # see where stretches of marker presence start and end
        marker_starts: dict[gt_marker.MarkerID,list[int]] = {}
        marker_ends  : dict[gt_marker.MarkerID,list[int]] = {}
        for m in markers:
            marker_starts[m], marker_ends[m] = gt_marker.get_appearance_starts_ends(markers[m], study_config.auto_code_episodes[e]['max_gap_duration'], study_config.auto_code_episodes[e]['min_duration'])
        # find potential interval starts and ends
        if len(study_config.auto_code_episodes[e]['start_markers'])>1:
            starts = gt_marker.get_sequence_interval(marker_starts, marker_ends, study_config.auto_code_episodes[e]['start_markers'], study_config.auto_code_episodes[e]['max_intermarker_gap_duration'], side='end')
        else:
            starts = marker_ends  [study_config.auto_code_episodes[e]['start_markers'][0]]
        if len(study_config.auto_code_episodes[e][ 'end_markers' ])>1:
            ends   = gt_marker.get_sequence_interval(marker_starts, marker_ends, study_config.auto_code_episodes[e][ 'end_markers' ], study_config.auto_code_episodes[e]['max_intermarker_gap_duration'], side='start')
        else:
            ends   = marker_starts[study_config.auto_code_episodes[e][ 'end_markers' ][0]]
        # now match interval starts and ends
        # strategy: run through starts and find latest start that is before first end (discard ends that are before the start)
        # keep pointer into array keeping track of ends and start that have already been discarded or consumed
        # NB: this assumes starts and ends are sorted, which the above procedures should indeed deliver
        intervals: list[tuple[int,int]] = []
        s_idx = 0
        e_idx = 0
        while s_idx<len(starts):
            # remove ends before the current start
            e_skip = np.nonzero(ends[e_idx:]<starts[s_idx])[0]
            if e_skip.size:
                e_idx += e_skip[-1]+1
            if e_idx > len(ends)-1:
                # we're out of ends, done
                break
            # for all starts in contention, find the last one that is before the next end
            gaps = ends[e_idx]-starts[s_idx:]
            # NB: it cannot occur that there are no starts before this end, since we move e_idx above in that case
            # and bail out if there are no ends left
            gaps[gaps<=0] = np.iinfo(gaps.dtype).max
            mini = np.argmin(gaps)
            # NB: intervals start the frame *after* the marker is last observed, and intervals end the frame *before* the marker is first observed
            intervals.append((starts[s_idx+mini]+1, ends[e_idx]-1))
            # these are consumed
            s_idx+=mini+1
            e_idx+=1
        # now insert into coding file. This just overwrites whatever is there
        episodes[e] = [y for x in intervals for y in x]

    # early exit if nothing has changed
    if episodes==episodes_original:
        if session.get_action_states(working_dir, True)[process.Action.AUTO_CODE_EPISODES]==process_pool.State.Completed:
            return
        session.update_action_states(working_dir, process.Action.AUTO_CODE_EPISODES, process_pool.State.Completed, study_config, unchanged=True)
        return

    # back up coding file if it exists
    if coding_file.is_file():
        shutil.move(coding_file, coding_file.with_stem(f'{naming.coding_file.split(".")[0]}_backup_before_episode_auto_code'))
    # store coded intervals to file
    episode.write_list_to_file(episode.marker_dict_to_list(episodes), coding_file)

    # update state
    session.update_action_states(working_dir, process.Action.AUTO_CODE_EPISODES, process_pool.State.Completed, study_config)


import pathlib
import shutil
import copy

from glassesTools import annotation, marker as gt_marker, process_pool

from .. import config, episode, naming, process, session


def run(working_dir: str|pathlib.Path, config_dir: str|pathlib.Path = None, **study_settings):
    working_dir = pathlib.Path(working_dir) # working directory of a session, not of a recording
    if config_dir is None:
        config_dir = config.guess_config_dir(working_dir)
    config_dir  = pathlib.Path(config_dir)
    print(f'processing: {working_dir.name}')

    # get settings for the study
    study_config = config.read_study_config_with_overrides(config_dir, {config.OverrideLevel.Session: working_dir.parent, config.OverrideLevel.Recording: working_dir}, **study_settings)
    if not study_config.auto_code_sync_points:
        raise ValueError(f'No automatic sync point detection is defined for this study, nothing to do')

    # get already coded interval(s), if any
    coding_file = working_dir / naming.coding_file
    if coding_file.is_file():
        episodes = episode.list_to_marker_dict(episode.read_list_from_file(coding_file), study_config.episodes_to_code)
        # flatten
        for e in episodes:
            episodes[e] = [i for iv in episodes[e] for i in iv]
    else:
        episodes = episode.get_empty_marker_dict(study_config.episodes_to_code)
    episodes_original = copy.deepcopy(episodes)

    # get marker files
    markers = [gt_marker.read_dataframe_from_file(m.m_id, m.aruco_dict_id, working_dir) for m in study_config.auto_code_sync_points['markers'] if gt_marker.get_file_name(m.m_id, m.aruco_dict_id, working_dir).is_file()]
    if not markers:
        missing_str = '\n- '.join([gt_marker.get_file_name(m.m_id, m.aruco_dict_id, None) for m in study_config.auto_code_sync_points['markers']])
        raise FileNotFoundError(f'None of the following marker files were found:\n- {missing_str}')
    # recode so we have a boolean with when markers are present
    markers = [gt_marker.code_for_presence(m, allow_failed=True) for m in markers if not m.empty]
    if not markers:
        raise RuntimeError(f'No markers found in the marker detection files for session "{working_dir.parent.name}", recording "{working_dir.name}"')
    # marker presence signal only contains marker detections (True). We need to fill the gaps in between detections with False (not detected) so we have a continuous signal without gaps
    markers = [gt_marker.expand_detection(m, fill_value=False) for m in markers]
    # see where stretches of True (marker presence) start
    marker_starts = [s for m in markers for s in gt_marker.get_appearance_starts_ends(m, study_config.auto_code_sync_points['max_gap_duration'], study_config.auto_code_sync_points['min_duration'])[0]]
    # insert in episodes
    [episodes[annotation.Event.Sync_Camera].append(i) for i in marker_starts if i not in episodes[annotation.Event.Sync_Camera]]

    # early exit if nothing has changed
    if episodes==episodes_original:
        if session.get_action_states(working_dir, True)[process.Action.AUTO_CODE_SYNC]==process_pool.State.Completed:
            return
        session.update_action_states(working_dir, process.Action.AUTO_CODE_SYNC, process_pool.State.Completed, study_config, unchanged=True)
        return

    # back up coding file if it exists
    if coding_file.is_file():
        shutil.move(coding_file, coding_file.with_stem(f'{naming.coding_file.split(".")[0]}_backup_before_sync_points_auto_code'))
    # store coded intervals to file
    episode.write_list_to_file(episode.marker_dict_to_list(episodes), coding_file)

    # update state
    session.update_action_states(working_dir, process.Action.AUTO_CODE_SYNC, process_pool.State.Completed, study_config)


import numpy as np
import cv2

from glassesTools import ocv

border_mask_cache: dict[tuple[int,int,float], np.ndarray] = {}

def detect_blob_HSV(proc_name: str, frame_idx: int, frame: np.ndarray, cam_params: ocv.CameraParams, low: tuple[int], high: tuple[int], edge_cut_fac=9.):
    # for instance:
    # low = (40, 60, 40)
    # high= (70, 255, 255)
    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2HSV)

    n_rows = frame.shape[0]
    n_cols = frame.shape[1]
    if (n_cols, n_rows, edge_cut_fac) not in border_mask_cache:
        border_pad_row = int(n_rows/edge_cut_fac)
        border_pad_col = int(n_cols/edge_cut_fac)

        border_mask = np.zeros(frame.shape[:2])
        border_mask[border_pad_row:(n_rows - border_pad_row), border_pad_col : (n_cols - border_pad_col)] = 1

        border_mask_cache[(n_cols, n_rows, edge_cut_fac)] = border_mask
    else:
        border_mask = border_mask_cache[(n_cols, n_rows, edge_cut_fac)]

    if low[0]>high[0]:
        mask = cv2.inRange(frame, (*low,), (180, *high[1:])) + cv2.inRange(frame, (0,*low[1:]), (*high,))
    else:
        mask = cv2.inRange(frame, (*low,), (*high,))

    # Consider targets only in center
    mask = mask * border_mask

    # Select largest blob
    output = cv2.connectedComponentsWithStats(mask.astype('uint8'), 8, cv2.CV_32S)
    labels = output[1]
    blob_sizes = output[2][:, -1]
    try:
        labels = labels == np.argsort(blob_sizes)[-2]
        mass_row, mass_col = np.where(labels)
        col = np.mean(mass_col)
        row = np.mean(mass_row)
    except:
        col = row = np.nan

    return col, row



import pathlib

from glassesTools import gaze_overlay_video, naming as gt_naming, process_pool, propagating_thread, timestamps
from glassesTools.gui import video_player

from .. import config, process, session


def run(working_dir: str|pathlib.Path, config_dir: str|pathlib.Path = None, show_visualization=False, progress_indicator: process_pool.JobProgress=None, **study_settings):
    # if show_visualization, each frame is shown in a viewer as video is generated
    working_dir = pathlib.Path(working_dir)
    if config_dir is None:
        config_dir = config.guess_config_dir(working_dir)
    config_dir  = pathlib.Path(config_dir)

    print(f'processing: {working_dir.parent.name}/{working_dir.name}')

    # if we need gui, we run processing in a separate thread (GUI needs to be on the main thread for OSX, see https://github.com/pthom/hello_imgui/issues/33)
    if show_visualization:
        gui = video_player.GUI(use_thread = False)
        gui.add_window(working_dir.name)
        gui.set_detachable(True)
        gui.set_show_controls(True)
        gui.set_show_play_percentage(True)
        gui.set_show_action_tooltip(True)
        gui.set_button_props_for_action(video_player.Action.Quit, 'Stop', tooltip='Interrupt (cut short) the video generation')

        proc_thread = propagating_thread.PropagatingThread(target=do_the_work, args=(working_dir, config_dir, gui, progress_indicator), kwargs=study_settings, cleanup_fun=gui.stop)
        proc_thread.start()
        gui.start()
        proc_thread.join()
    else:
        do_the_work(working_dir, config_dir, None, progress_indicator, **study_settings)


def do_the_work(working_dir: pathlib.Path, config_dir: pathlib.Path, gui: video_player.GUI, progress_indicator: process_pool.JobProgress, **study_settings):
    # progress indicator
    if progress_indicator is None:
        progress_indicator = process_pool.JobProgress(printer=lambda x: print(x))
    progress_indicator.set_unit('frames')
    progress_indicator.set_start_time_to_now()

    # get settings for the study
    study_config = config.read_study_config_with_overrides(config_dir, {config.OverrideLevel.Session: working_dir.parent, config.OverrideLevel.Recording: working_dir}, **study_settings)

    # get info about recording
    rec_def = study_config.session_def.get_recording_def(working_dir.name)
    if rec_def.type!=session.RecordingType.Eye_Tracker:
        raise ValueError(f'You can only run gaze_overlay_video on eye tracker recordings, not on a {str(rec_def.type).split(".")[1]} recording')
    in_video = session.read_recording_info(working_dir, rec_def.type)[1]
    video_ts = timestamps.VideoTimestamps(working_dir / gt_naming.frame_timestamps_fname)

    # set up gaze overlay video maker and run it
    video_maker = gaze_overlay_video.VideoMaker(working_dir, in_video, video_ts, working_dir / gt_naming.scene_camera_calibration_fname, working_dir / gt_naming.gaze_data_fname)
    video_maker.set_vid_pos_look(study_config.overlay_video_gaze_vid_pos_color, study_config.overlay_video_gaze_vid_pos_radius, study_config.overlay_video_gaze_vid_pos_thickness)
    video_maker.set_world_pos_look(study_config.overlay_video_gaze_world_pos_color, study_config.overlay_video_gaze_world_pos_radius, study_config.overlay_video_gaze_world_pos_thickness)
    video_maker.attach_gui(gui)
    if gui is not None:
        gui.set_show_timeline(True, video_ts, window_id=gui.main_window_id)

    # prep progress indicator
    total = video_ts.get_last()[0]
    progress_indicator.set_total(total)
    progress_indicator.set_intervals(int(total/200), int(total/200))

    # update state: set to not run so that if we crash or cancel below the task is correctly marked as not run (video files are corrupt)
    session.update_action_states(working_dir, process.Action.MAKE_GAZE_OVERLAY_VIDEO, process_pool.State.Not_Run, study_config)

    # now run
    video_maker.process_video(progress_indicator.update)

    # update state
    session.update_action_states(working_dir, process.Action.MAKE_GAZE_OVERLAY_VIDEO, process_pool.State.Completed, study_config)



import pandas as pd
import numpy as np
from typing import Any

from glassesTools import data_files


def insert_ts_fridx_in_df(df: pd.DataFrame, object: Any, suffix: str, ts: np.ndarray, fridxs: np.ndarray):
    # figure out where
    cols = [c for cs in data_files.uncompress_columns(object._columns_compressed) for c in cs if c in df.columns or c in [f'timestamp_{suffix}',f'frame_idx_{suffix}']]
    ts_idx = cols.index(f'timestamp_{suffix}')
    fr_idx = cols.index(f'frame_idx_{suffix}')
    # figure out what order
    cols = ['timestamp','frame_idx']
    if ts_idx>fr_idx:
        cols = [cols[1], cols[0]]
    # insert
    for c in cols:
        v = ts if c=='timestamp' else fridxs
        i = ts_idx if c=='timestamp' else fr_idx
        if f'{c}_{suffix}' in df.columns:
            df[f'{c}_{suffix}'] = v
        else:
            df.insert(i, f'{c}_{suffix}', v)

    return df



import enum
import typing

from glassesTools import annotation, json, process_pool, utils

class Action(enum.IntEnum):
    IMPORT = enum.auto()
    MAKE_GAZE_OVERLAY_VIDEO = enum.auto()
    CODE_EPISODES = enum.auto()
    DETECT_MARKERS = enum.auto()
    GAZE_TO_PLANE = enum.auto()
    AUTO_CODE_SYNC = enum.auto()
    AUTO_CODE_EPISODES = enum.auto()
    SYNC_ET_TO_CAM = enum.auto()
    SYNC_TO_REFERENCE = enum.auto()
    VALIDATE = enum.auto()
    EXPORT_TRIALS = enum.auto()
    MAKE_MAPPED_GAZE_VIDEO = enum.auto()
    @property
    def displayable_name(self):
        return self.name.replace("_", " ").title()
    @property
    def needs_GUI(self):
        return self in [Action.CODE_EPISODES, Action.SYNC_ET_TO_CAM]
    @property
    def has_options(self):
        return self in [Action.MAKE_GAZE_OVERLAY_VIDEO, Action.DETECT_MARKERS, Action.GAZE_TO_PLANE, Action.MAKE_MAPPED_GAZE_VIDEO]
    @property
    def has_progress_indication(self):
        return self in [Action.MAKE_GAZE_OVERLAY_VIDEO, Action.DETECT_MARKERS, Action.GAZE_TO_PLANE, Action.VALIDATE, Action.MAKE_MAPPED_GAZE_VIDEO]
    def succ(self):
        v = self.value+1
        if v > Action.MAKE_MAPPED_GAZE_VIDEO.value:
            raise StopIteration('Enumeration ended')
        return Action(v)
    def pred(self):
        v = self.value-1
        if v<Action.IMPORT.value:
            raise StopIteration('Enumeration ended')
        return Action(v)
    def next_values(self, inclusive=False) -> set['Action']:
        vals: set[Action] = set()
        a = self
        if inclusive:
            vals.add(a)
        try:
            while True:
                a = a.succ()
                vals.add(a)
        except StopIteration:
            pass    # we're done
        return vals
def action_str_to_enum_val(x: str) -> Action:
    return utils.enum_str_2_val(x, Action, {'MAKE_VIDEO':'MAKE_MAPPED_GAZE_VIDEO', 'RUN_VALIDATION':'VALIDATE', 'AUTO_CODE_TRIALS': 'AUTO_CODE_EPISODES'})
json.register_type(json.TypeEntry(Action, '__enum.process.Action__', utils.enum_val_2_str, action_str_to_enum_val))

def action_to_func(action: Action) -> typing.Callable[..., None]:
    # Returns function to perform the provided action. NB: not for Action.IMPORT,
    # needs its own special handling by caller instead
    from .make_gaze_overlay_video import run as make_gaze_overlay_video
    from .code_episodes import run as do_coding
    from .detect_markers import run as detect_markers
    from .sync_et_to_cam import run as sync_et_to_cam
    from .sync_to_ref import run as sync_to_ref
    from .gaze_to_plane import run as gaze_to_plane
    from .export_trials import run as export_trials
    from .run_validation import run as run_validation
    from .auto_code_sync_points import run as auto_code_sync_points
    from .auto_code_episodes import run as auto_code_episodes
    from .make_mapped_gaze_video import run as make_mapped_gaze_video

    match action:
        case Action.IMPORT:
            return None # Needs a special case handled by the caller
        case Action.MAKE_GAZE_OVERLAY_VIDEO:
            return make_gaze_overlay_video
        case Action.CODE_EPISODES:
            return do_coding
        case Action.DETECT_MARKERS:
            return detect_markers
        case Action.GAZE_TO_PLANE:
            return gaze_to_plane
        case Action.AUTO_CODE_SYNC:
            return auto_code_sync_points
        case Action.AUTO_CODE_EPISODES:
            return auto_code_episodes
        case Action.SYNC_ET_TO_CAM:
            return sync_et_to_cam
        case Action.SYNC_TO_REFERENCE:
            return sync_to_ref
        case Action.VALIDATE:
            return run_validation
        case Action.EXPORT_TRIALS:
            return export_trials
        case Action.MAKE_MAPPED_GAZE_VIDEO:
            return make_mapped_gaze_video
        case _:
            raise NotImplementedError(f'Logic is not implemented for {action.displayable_name} ({action}), major developer oversight! Let him know.')

def is_session_level_action(action: Action) -> bool:
    return action in [Action.SYNC_TO_REFERENCE, Action.EXPORT_TRIALS, Action.MAKE_MAPPED_GAZE_VIDEO]

def is_action_possible_given_config(action: Action, study_config: 'config.Study') -> bool:
    match action:
        case Action.CODE_EPISODES | Action.DETECT_MARKERS | Action.GAZE_TO_PLANE:
            # if there is any form of config beyond a session definition, these will be possible (whether config is complete is handled by error messages so no need to worry about that here)
            return not not study_config.planes or not not study_config.episodes_to_code or not not study_config.planes_per_episode or not not study_config.individual_markers
        case Action.AUTO_CODE_SYNC:
            return not not study_config.auto_code_sync_points
        case Action.AUTO_CODE_EPISODES:
            return not not study_config.auto_code_episodes
        case Action.SYNC_ET_TO_CAM:
            return study_config.get_cam_movement_for_et_sync_method in ['plane', 'function']
        case Action.SYNC_TO_REFERENCE:
            return not not study_config.sync_ref_recording
        case Action.VALIDATE:
            return annotation.Event.Validate in study_config.planes_per_episode
        case Action.EXPORT_TRIALS:
            return True     # always possible (in terms of config), since gaze overlay videos are always possible
        case Action.MAKE_MAPPED_GAZE_VIDEO:
            return not not study_config.mapped_video_make_which

        case _:
            # no config preconditions for the other actions
            return True

def is_action_possible_for_recording(rec: str, rec_type: 'session.RecordingType', action: Action, study_config: 'config.Study') -> bool:
    from .. import session
    if rec_type==session.RecordingType.Camera and action in [Action.MAKE_GAZE_OVERLAY_VIDEO, Action.GAZE_TO_PLANE, Action.SYNC_ET_TO_CAM, Action.VALIDATE]:
        return False
    elif action==Action.AUTO_CODE_EPISODES and study_config.sync_ref_recording and rec!=study_config.sync_ref_recording:
        # if we have a sync_ref_recording, automatic coding of trials is only possible for the sync_ref_recording, not the other recordings
        return False
    return True

def get_actions_for_config(study_config: 'config.Study', exclude_session_level: bool=False) -> set[Action]:
    actions = {a for a in Action if is_action_possible_given_config(a, study_config)}
    if exclude_session_level:
        actions = {a for a in actions if not is_session_level_action(a)}
    return actions

def _determine_to_invalidate(action: Action, study_config: 'config.Study') -> tuple[set[Action],bool]:
    for_all_recs = False    # if True, recording-level states should be invalidated for all recordings in a session
    match action:
        case Action.IMPORT:
            states_to_invalidate = action.next_values()
        case Action.MAKE_GAZE_OVERLAY_VIDEO:
            states_to_invalidate = {Action.EXPORT_TRIALS}
        case Action.CODE_EPISODES:
            states_to_invalidate = {a for a in action.next_values() if a not in [Action.AUTO_CODE_SYNC, Action.AUTO_CODE_EPISODES]}
            for_all_recs = not not study_config.sync_ref_recording
        case Action.DETECT_MARKERS:
            # N.B.: don't invalidate auto code sync and auto code episodes as the individual markers used for these
            # automated coding actions are always detected throughout the whole video
            actions = {a for a in action.next_values() if a not in [Action.SYNC_TO_REFERENCE, Action.AUTO_CODE_SYNC, Action.AUTO_CODE_EPISODES]}
            if (study_config.mapped_video_process_planes_for_all_frames or
                study_config.mapped_video_plane_marker_color is not None or # NB: no need to check mapped_video_recovered_plane_marker_color as it only overrides colors for some plane markers
                study_config.mapped_video_individual_marker_color is not None or
                study_config.mapped_video_unexpected_marker_color is not None or
                study_config.mapped_video_rejected_marker_color is not None):
                # in this case MAKE_MAPPED_GAZE_VIDEO processes each frame itself, so output of DETECT_MARKERS is not used
                actions.discard(Action.MAKE_MAPPED_GAZE_VIDEO)
            states_to_invalidate = actions
        case Action.GAZE_TO_PLANE:
            # NB: SYNC_ET_TO_CAM and SYNC_TO_REFERENCE operate on gazeData not gaze on plane data, and MAKE_VIDEO does the job itself/doesn't use this file
            states_to_invalidate = {a for a in action.next_values() if a not in [Action.AUTO_CODE_SYNC, Action.AUTO_CODE_EPISODES, Action.SYNC_ET_TO_CAM, Action.SYNC_TO_REFERENCE, Action.MAKE_MAPPED_GAZE_VIDEO]}
        case Action.AUTO_CODE_SYNC:
            states_to_invalidate = {Action.CODE_EPISODES, Action.GAZE_TO_PLANE, Action.SYNC_TO_REFERENCE, Action.EXPORT_TRIALS, Action.MAKE_MAPPED_GAZE_VIDEO}
        case Action.AUTO_CODE_EPISODES:
            states_to_invalidate = {a for a in Action.CODE_EPISODES.next_values(inclusive=True) if a not in [Action.AUTO_CODE_EPISODES, Action.AUTO_CODE_SYNC, Action.SYNC_ET_TO_CAM, Action.SYNC_TO_REFERENCE, Action.VALIDATE]}
            for_all_recs = not not study_config.sync_ref_recording and annotation.Event.Trial in study_config.auto_code_episodes
        case Action.SYNC_ET_TO_CAM:
            states_to_invalidate = {a for a in Action.GAZE_TO_PLANE.next_values(inclusive=True) if a not in [Action.AUTO_CODE_EPISODES, Action.AUTO_CODE_SYNC, Action.SYNC_ET_TO_CAM]}
        case Action.SYNC_TO_REFERENCE:
            states_to_invalidate = {a for a in Action.GAZE_TO_PLANE.next_values(inclusive=True) if a not in [Action.AUTO_CODE_EPISODES, Action.AUTO_CODE_SYNC, Action.SYNC_ET_TO_CAM, Action.SYNC_TO_REFERENCE]}
        case Action.VALIDATE:
            states_to_invalidate = {Action.EXPORT_TRIALS}
        case Action.EXPORT_TRIALS:
            states_to_invalidate = set()
        case Action.MAKE_MAPPED_GAZE_VIDEO:
            states_to_invalidate = set()
        case _:
            raise NotImplementedError(f'Logic is not implemented for {action.displayable_name} ({action}), major developer oversight! Let him know.')
    return states_to_invalidate, for_all_recs

def action_update_and_invalidate(action: Action, state: process_pool.State, study_config: 'config.Study') -> tuple[dict[Action, process_pool.State], bool]:
    # set status of indicated task
    action_state_mutations = {action: state}

    # determine what other (later) actions are invalidated (and should thus be reset to not started state) by this action
    # being performed. There may be a better way of doing this, but i prefer to actively, per case, think this through
    # and explicitly write it out
    states_to_invalidate, for_all_recs = _determine_to_invalidate(action, study_config)
    for a in states_to_invalidate:
        action_state_mutations[a] = process_pool.State.Not_Run

    return action_state_mutations, for_all_recs

def _is_recording_action_possible(rec: str, action_states: dict[Action, process_pool.State], study_config: 'config.Study', rec_type: 'session.RecordingType', action: Action) -> tuple[bool,list[Action]]:
    if not is_action_possible_given_config(action, study_config):
        return False, None
    elif not is_action_possible_for_recording(rec, rec_type, action, study_config):
        return False, None

    preconditions = {Action.IMPORT} # IMPORT is a precondition for all actions except IMPORT itself
    match action:
        case Action.IMPORT:
            return action_states[Action.IMPORT]==process_pool.State.Not_Run, []  # possible if not already imported
        case Action.MAKE_GAZE_OVERLAY_VIDEO:
            pass    # nothing besides import
        case Action.CODE_EPISODES:
            pass    # nothing besides import
        case Action.DETECT_MARKERS:
            if not (study_config.auto_code_sync_points or study_config.auto_code_episodes):
                preconditions.add(Action.CODE_EPISODES)
        case Action.GAZE_TO_PLANE:
            preconditions.update([Action.CODE_EPISODES, Action.DETECT_MARKERS])
            if study_config.sync_ref_recording:
                preconditions.add(Action.SYNC_TO_REFERENCE)
            if study_config.get_cam_movement_for_et_sync_method!='':
                preconditions.add(Action.SYNC_ET_TO_CAM)
        case Action.AUTO_CODE_SYNC:
            preconditions.update([Action.DETECT_MARKERS])
        case Action.AUTO_CODE_EPISODES:
            preconditions.update([Action.DETECT_MARKERS])
        case Action.SYNC_ET_TO_CAM:
            preconditions.update([Action.CODE_EPISODES, Action.DETECT_MARKERS])
        case Action.VALIDATE:
            preconditions.update([Action.CODE_EPISODES, Action.GAZE_TO_PLANE])
        case _:
            raise NotImplementedError(f'Logic is not implemented for {action.displayable_name} ({action}), major developer oversight! Let him know.')

    # check that preconditions are met
    states = [action_states[p]==process_pool.State.Completed for p in preconditions]
    missing = [p for p,s in zip(preconditions,states) if not s]
    return all(states), [p for p in Action if p in missing] # NB: ensure stable order

def _is_session_action_possible(session_action_states: dict[Action, process_pool.State], recording_action_states: dict[str,dict[Action, process_pool.State]], study_config: 'config.Study', rec_types: dict[str,'session.RecordingType'], action: Action) -> tuple[bool, list[Action]]:
    if not is_action_possible_given_config(action, study_config):
        return False, None

    preconditions = set()
    preconditions_test = 'and'
    match action:
        case Action.SYNC_TO_REFERENCE:
            preconditions.update([Action.CODE_EPISODES])
            # NB: even if study_config.auto_code_sync_points is defined, user may decide to code sync manually. So don't add Action.AUTO_CODE_SYNC to preconditions
            if study_config.get_cam_movement_for_et_sync_method!='':
                preconditions.add(Action.SYNC_ET_TO_CAM)
        case Action.EXPORT_TRIALS:
            preconditions.update([Action.MAKE_GAZE_OVERLAY_VIDEO, Action.GAZE_TO_PLANE, Action.VALIDATE, Action.MAKE_MAPPED_GAZE_VIDEO])
            preconditions_test = 'or'
        case Action.MAKE_MAPPED_GAZE_VIDEO:
            preconditions.update([Action.CODE_EPISODES])
            # NB: even if study_config.auto_code_sync_points is defined, user may decide to code sync manually. So don't add Action.AUTO_CODE_SYNC to preconditions
            # NB: even if study_config.auto_code_episodes is defined, user may decide to code episodes manually. So don't add Action.AUTO_CODE_EPISODES to preconditions
            if study_config.sync_ref_recording:
                preconditions.add(Action.SYNC_TO_REFERENCE)
            elif study_config.get_cam_movement_for_et_sync_method!='':
                preconditions.add(Action.SYNC_ET_TO_CAM)
        case _:
            raise NotImplementedError(f'Logic is not implemented for {action.displayable_name} ({action}), major developer oversight! Let him know.')

    precond_met: dict[Action,tuple[bool,list[str]]] = {}
    for p in [p for p in Action if p in preconditions]: # NB: ensure stable order
        if is_session_level_action(p):
            met = session_action_states[p]==process_pool.State.Completed
            precond_met[p] = (met,None)
        else:
            met1 = [is_action_possible_for_recording(r, rec_types[r], p, study_config) for r in recording_action_states]
            met2 = [recording_action_states[r][p]==process_pool.State.Completed for r in recording_action_states]
            met = [(not m1) or m2 for m1,m2 in zip(met1,met2)]    # not m1 because ignore if action isn't possible for that recording anyway
            all_met = all(met) if met else False
            precond_met[p] = (all_met, [] if all_met else [r for r,m1,m2 in zip(recording_action_states,met1,met2) if m1 and not m2])

    if preconditions_test=='and':
        ok = all((precond_met[p][0] for p in precond_met))
    else:
        ok = any((precond_met[p][0] for p in precond_met))
    return ok, {p: precond_met[p][1] for p in precond_met if not precond_met[p][0]} if not ok else {}

def get_possible_actions(session_action_states: dict[Action, process_pool.State], recording_action_states: dict[str,dict[Action, process_pool.State]], actions_to_check: set[Action], study_config: 'config.Study') -> dict[Action, tuple[bool|list[str], dict[Action,list[str]]]]:
    # determine based on actions_states which actions have all their preconditions met. Return a set containing just
    # those possible actions
    # actions_to_check can be a subset of all actions, if user e.g. knows some actions aren't possible or wanted due to settings
    # this function doesn't check that
    merged_states = {r:session_action_states|recording_action_states[r] for r in recording_action_states}
    rec_types = {r.name:r.type for r in study_config.session_def.recordings}

    possible_actions: dict[Action,tuple[bool|list[str],dict[Action,list[str]]]] = {}  # per action list if possible/recordings for which its possible, and also which preconditions are unmet
    for a in actions_to_check:
        if is_session_level_action(a):
            states = _is_session_action_possible(session_action_states, recording_action_states, study_config, rec_types, a)
            if states[0] or states[1]:
                possible_actions[a] = states
        else:
            states = {r:_is_recording_action_possible(r, merged_states[r], study_config, rec_types[r], a) for r in merged_states}
            possible_recs = [r for r in merged_states if states[r][0]]
            fails = _get_precond_fails_for_rec(states)
            if possible_recs or fails:
                possible_actions[a] = possible_recs, fails

    return possible_actions

def _get_precond_fails_for_rec(states: dict[str,tuple[bool,list[Action]]]) -> dict[Action,None|list[str]]:
    out: dict[Action,None|list[str]] = {}
    for r in states:
        if states[r][0] or states[r][1] is None:
            continue
        for a in states[r][1]:
            if is_session_level_action(a):
                out[a] = None
            elif a not in out:
                out[a] = [r]
            else:
                out[a].append(r)
    return out

def get_impossible_reason_text(action: Action, states: dict[str, dict[Action,tuple[bool|list[str],dict[Action,list[str]]]]], study_config: 'config.Study', for_recording=False, for_single=False) -> str:
    reason = f'Cannot run {action.displayable_name} because required actions have not yet been run'
    if for_single:
        reason += f':\n{_get_impossible_reason_text_impl(states[action][1], "", study_config)}'
    else:
        reason += ' for the following '
        if for_recording:
            reason += 'recording(s):\n'
        else:
            reason += 'session(s):\n'
        for s in states:
            if action not in states[s]:
                continue
            ac = states[s][action][1]
            reason += f'- {s}:\n{_get_impossible_reason_text_impl(ac, s, study_config)}'
    return reason

def _get_impossible_reason_text_impl(fails: dict[Action,list[str]], ref: str, study_config: 'config.Study') -> str:
    reason = ''
    for a in fails:
        if not is_action_possible_given_config(a, study_config):
            continue
        reason += f'  - {a.displayable_name}'
        if not not fails[a] and (len(fails[a])>1 or fails[a][0]!=ref):
            reason += f' ({", ".join(fails[a])})'
        reason += '\n'
    return reason


import pathlib
import typing
import numpy as np
import cv2

from glassesTools import aruco

from . import image_helper
from ... import config, process, session


def is_project_folder(path: str | pathlib.Path):
    path = pathlib.Path(path)
    if not path.is_dir():
        return False
    # a project directory should contain a 'config'
    # folder and inside that are 'study_def.json' and 'session_def.json' file
    return (path/'config').is_dir() and \
        (path/'config'/config.Study.default_json_file_name).is_file() and \
        (path/'config'/session.SessionDefinition.default_json_file_name).is_file()

def init_project_folder(path: str | pathlib.Path):
    path = pathlib.Path(path)
    if not path.is_dir():
        raise ValueError(f'The provided path is not a folder. Provided path: {path}')
    config_dir = path/'config'
    if not config_dir.is_dir():
        config_dir.mkdir()
    # make empty 'study_def.json' file
    study_config = config.Study.get_empty(path)
    study_config.store_as_json(config_dir)


def get_aruco_marker_image(sz: int, m_id: int, dictionary_id: int, marker_border_bits: int):
    marker_image = aruco.get_marker_image(sz, m_id, dictionary_id, marker_border_bits)
    if marker_image is None:
        return None
    return image_helper.ImageHelper(marker_image)

def load_image_with_helper(path_or_image: pathlib.Path|np.ndarray):
    if isinstance(path_or_image, pathlib.Path):
        return image_helper.ImageHelper(cv2.cvtColor(cv2.imread(path_or_image, cv2.IMREAD_COLOR),cv2.COLOR_BGR2RGB))
    else:
        return image_helper.ImageHelper(path_or_image)


class JobInfo(typing.NamedTuple):
    action:     process.Action
    session:    str
    recording:  typing.Optional[str] = None


import pathlib
import typing
import shutil
import os
import asyncio
import subprocess
import pathvalidate
import threading
from imgui_bundle import imgui, imspinner, hello_imgui, icons_fontawesome_6 as ifa6

from glassesTools import annotation, aruco, async_thread, camera_recording, eyetracker, gui as gt_gui, marker as gt_marker, naming as gt_naming, platform, process_pool, recording, video_utils
from glassesTools.validation import config as val_config, default_poster, DataQualityType, dynamic as val_dynamic, export, get_DataQualityType_explanation
from glassesTools.gui import colors

from . import utils
from ... import config, marker, naming, plane, process, session

_picker_last_folder = None
def get_folder_picker(g, reason: str, *args, **kwargs):
    from . import gui
    g = typing.cast(gui.GUI,g)  # indicate type to typechecker
    def select_callback(selected: list[pathlib.Path]):
        global _picker_last_folder
        if len(selected)==1 and selected[0].is_dir():
            _picker_last_folder = selected[0]
        else:
            _picker_last_folder = selected[0].parent
        match reason:
            case 'loading' | 'creating':
                try_load_project(g, selected, action=reason)
            case 'add_et_recordings':
                add_eyetracking_recordings(g, selected, *args, **kwargs)
            case 'add_cam_recordings':
                camera_show_glob_filter_config(g, selected, *args, **kwargs)
            case 'set_default_cam_cal':
                set_default_cam_cal(selected[0], *args, **kwargs)
            case 'set_cam_cal':
                set_cam_cal(selected[0], *args, **kwargs)
            case 'deploy_aruco':
                aruco.deploy_marker_images(selected[0], 1000, *args, **kwargs)
            case 'deploy_gv_poster_pdf':
                default_poster.deploy_default_pdf(selected[0])
            case 'export':
                show_export_config(g, selected[0], *args, **kwargs)
            case 'new_plane_dyn_dir':
                kwargs['callback'](selected[0])
            case 'deploy_gv_dynamic_script':
                not_copied = val_dynamic.deploy_setup_and_script(selected[0])
                if not_copied:
                    buttons = {
                        ifa6.ICON_FA_CHECK+" Yes": lambda: val_dynamic.deploy_setup_and_script(selected[0], overwrite=True),
                        ifa6.ICON_FA_CIRCLE_XMARK+" No": None
                    }
                    files = '- '+('\n- '.join(not_copied))
                    gt_gui.utils.push_popup(g, gt_gui.msg_box.msgbox, "Overwrite files", f"The following files were not deployed as they already exist\nin the selected folder:\n{files}\n\nDo you want to overwrite these files?", gt_gui.msg_box.MsgBox.question, buttons)
            case 'gv_dynamic_import_setup':
                set_dynamic_validation_plane_config(g, selected[0], **kwargs)
            case _:
                raise ValueError(f'reason "{reason}" not understood')

    match reason:
        case 'loading' | 'creating':
            header = "Select or drop project folder"
            allow_multiple = False
            picker_type = gt_gui.file_picker.DirPicker
        case 'add_et_recordings':
            header = "Select or drop recording folders"
            allow_multiple = True
            picker_type = gt_gui.file_picker.DirPicker
        case 'add_cam_recordings':
            header = "Select or drop recording folders or files"
            allow_multiple = True
            picker_type = gt_gui.file_picker.FilePicker
        case 'set_default_cam_cal' | 'set_cam_cal':
            header = "Select or drop calibration xml file"
            allow_multiple = False
            picker_type = gt_gui.file_picker.FilePicker
        case 'deploy_aruco':
            header = "Select folder to store ArUco marker images"
            allow_multiple = False
            picker_type = gt_gui.file_picker.DirPicker
        case 'deploy_gv_poster_pdf':
            header = "Select folder to store default glassesValidator poster pdf"
            allow_multiple = False
            picker_type = gt_gui.file_picker.DirPicker
        case 'export':
            header = "Select folder to store results export"
            allow_multiple = False
            picker_type = gt_gui.file_picker.DirPicker
        case 'new_plane_dyn_dir' | 'gv_dynamic_import_setup':
            header = "Select or drop setup json file for dynamic validation PsychoPy script"
            allow_multiple = False
            picker_type = gt_gui.file_picker.FilePicker
        case 'deploy_gv_dynamic_script':
            header = "Select folder to store dynamic validation PsychoPy script and default configuration"
            allow_multiple = False
            picker_type = gt_gui.file_picker.DirPicker
        case _:
            raise ValueError(f'reason "{reason}" not understood')
    picker = picker_type(title=header, allow_multiple=allow_multiple, callback=select_callback, start_dir=_picker_last_folder)
    picker.set_show_only_dirs(False)
    return picker

def try_load_project(g, path: str|pathlib.Path, action='loading'):
    from . import gui
    g = typing.cast(gui.GUI,g)  # indicate type to typechecker
    if isinstance(path,list):
        if not path:
            gt_gui.utils.push_popup(g, gt_gui.msg_box.msgbox, "Project opening error", "A single project directory should be provided. None provided so cannot open.", gt_gui.msg_box.MsgBox.error, more="Dropped paths:\n"+('\n'.join([str(p) for p in path])))
            return
        elif len(path)>1:
            gt_gui.utils.push_popup(g, gt_gui.msg_box.msgbox, "Project opening error", f"Only a single project directory should be provided, but {len(path)} were provided. Cannot open multiple projects.", gt_gui.msg_box.MsgBox.error, more="Dropped paths:\n"+('\n'.join([str(p) for p in path])))
            return
        else:
            path = path[0]
    path = pathlib.Path(path)

    if utils.is_project_folder(path):
        if action=='creating':
            buttons = {
                ifa6.ICON_FA_CHECK+" Yes": lambda: g.load_project(path),
                ifa6.ICON_FA_CIRCLE_XMARK+" No": None
            }
            gt_gui.utils.push_popup(g, gt_gui.msg_box.msgbox, "Create new project", "The selected folder is already a project folder.\nDo you want to open it?", gt_gui.msg_box.MsgBox.question, buttons)
        else:
            g.load_project(path)
    elif any(path.iterdir()):
        if action=='creating':
            gt_gui.utils.push_popup(g, gt_gui.msg_box.msgbox, "Project creation error", "The selected folder is not empty. Cannot be used to create a project folder.", gt_gui.msg_box.MsgBox.error)
        else:
            gt_gui.utils.push_popup(g, gt_gui.msg_box.msgbox, "Project opening error", "The selected folder is not a project folder. Cannot open.", gt_gui.msg_box.MsgBox.error)
    else:
        def init_project_and_ask():
            utils.init_project_folder(path)
            buttons = {
                ifa6.ICON_FA_CHECK+" Yes": lambda: g.load_project(path),
                ifa6.ICON_FA_CIRCLE_XMARK+" No": None
            }
            gt_gui.utils.push_popup(g, gt_gui.msg_box.msgbox, "Open new project", "Do you want to open the new project folder?", gt_gui.msg_box.MsgBox.question, buttons)
        if action=='creating':
            init_project_and_ask()
        else:
            buttons = {
                ifa6.ICON_FA_CHECK+" Yes": lambda: init_project_and_ask(),
                ifa6.ICON_FA_CIRCLE_XMARK+" No": None
            }
            gt_gui.utils.push_popup(g, gt_gui.msg_box.msgbox, "Create new project", "The selected folder is empty. Do you want to use it as a new project folder?", gt_gui.msg_box.MsgBox.warn, buttons)

def set_dynamic_validation_plane_config(g, config_file_path: pathlib.Path, p_def: plane.Definition_GlassesValidator):
    from . import gui
    g = typing.cast(gui.GUI,g)  # indicate type to typechecker
    path = config.guess_config_dir(g.study_config.working_directory)
    p_dir = path / p_def.name
    try:
        if config_file_path.suffix!='.json':
            raise ValueError('The selected file is not a json file. Please select a json file.')
        auto_coding_setup = val_dynamic.setup_to_plane_config(p_dir, config_file_path.parent, config_file_path.name)
    except Exception as exc:
        gt_gui.utils.push_popup(g, gt_gui.msg_box.msgbox, "Error converting setup", f"There was an error trying to load the dynamic validation setup:\n{exc}", gt_gui.msg_box.MsgBox.error, more=gt_gui.utils.get_traceback(type(exc), exc, exc.__traceback__))
        return
    # N.B.: all we have to do here for the plane. File watcher will catch the change and reload the plane
    set_auto_coding_for_dynamic_validation_plane(g, p_def, auto_coding_setup)

def set_auto_coding_for_dynamic_validation_plane(g, p_def: plane.Definition_GlassesValidator, auto_coding_setup: dict[str,list[gt_marker.MarkerID]]):
    def _apply_individual_markers_if_needed():
        all_markers = set(g.study_config.auto_code_episodes[annotation.Event.Validate]['start_markers'])
        all_markers.update(g.study_config.auto_code_episodes[annotation.Event.Validate]['end_markers'])
        applied = False
        for m in all_markers:
            if not any((im.id==m.m_id and im.aruco_dict_id==m.aruco_dict_id for im in g.study_config.individual_markers)):
                g.study_config.individual_markers.append(marker.Marker(m.m_id, detect_only=True, aruco_dict_id=m.aruco_dict_id, marker_border_bits=auto_coding_setup['marker_border_bits']))
                applied = True
        return applied
    def _apply_automatic_val_coding():
        if not g.study_config.auto_code_episodes:
            g.study_config.auto_code_episodes = {}
        g.study_config.auto_code_episodes[annotation.Event.Validate] = config.AutoCodeEpisodes(start_markers=auto_coding_setup['start_markers'], end_markers=auto_coding_setup['end_markers'])
        g.study_config.auto_code_episodes[annotation.Event.Validate].apply_defaults()
        _apply_individual_markers_if_needed()
        # persist changed config
        g.study_config.store_as_json()

    from . import gui
    g = typing.cast(gui.GUI,g)  # indicate type to typechecker
    exists = g.study_config.auto_code_episodes and annotation.Event.Validate in g.study_config.auto_code_episodes
    if exists:
        matches = all((g.study_config.auto_code_episodes[annotation.Event.Validate][f]==auto_coding_setup[f] for f in ('start_markers','end_markers')))
        if matches:
            # nothing to do, do check all individual markers are there
            if _apply_individual_markers_if_needed():
                # persist changed config
                g.study_config.store_as_json()
            return
    # ask user if they want to apply settings
    buttons = {
        ifa6.ICON_FA_CHECK+" Yes": lambda: _apply_automatic_val_coding(),
        ifa6.ICON_FA_CIRCLE_XMARK+" No": None
    }
    msg = f'The following setup for automatic coding of {annotation.tooltip_map[annotation.Event.Validate]}s was found upon import of the dynamic validation setup for the {p_def.name} plane:\n'
    for f in ('start_markers','end_markers'):
        msg += f'{f}: ['+(', '.join((gt_marker.marker_ID_to_str(m) for m in auto_coding_setup[f])))+']\n'
    if exists:
        msg += f'\nThe following setup for automatic coding of {annotation.tooltip_map[annotation.Event.Validate]}s is already set:\n'
        for f in ('start_markers','end_markers'):
            msg += f'{f}: ['+(', '.join((gt_marker.marker_ID_to_str(m) for m in g.study_config.auto_code_episodes[annotation.Event.Validate][f])))+']\n'
        msg += '\nDo you want to overwrite this configuration with the above?'
    else:
        msg += '\nDo you want to apply this configuration?'
    gt_gui.utils.push_popup(g, gt_gui.msg_box.msgbox, ("Overwrite" if exists else "Apply")+f" automated {annotation.tooltip_map[annotation.Event.Validate]} coding config?", msg, gt_gui.msg_box.MsgBox.question, buttons)

def set_default_cam_cal(cal_path: str|pathlib.Path, rec_def: session.RecordingDefinition, rec_def_path: pathlib.Path):
    rec_def.set_default_cal_file(cal_path, rec_def_path)

def set_cam_cal(cal_path: str|pathlib.Path, working_directory: str|pathlib.Path):
    cal_path = pathlib.Path(cal_path)
    working_directory = pathlib.Path(working_directory)
    shutil.copyfile(str(cal_path), str(working_directory / gt_naming.scene_camera_calibration_fname))

def delete_cam_cal(working_directory: str|pathlib.Path):
    pathlib.Path(working_directory / gt_naming.scene_camera_calibration_fname).unlink(missing_ok=True)

def make_plane(g, p_type: plane.Type, name: str, is_dynamic: bool, dynamic_config_file: str):
    from . import gui
    g = typing.cast(gui.GUI,g)  # indicate type to typechecker
    path = config.guess_config_dir(g.study_config.working_directory)
    p_dir = path / name
    # deal with dynamic glassesValidator plane
    extra = {}
    if is_dynamic:
        extra['use_default'] = not dynamic_config_file
        extra['is_dynamic'] = True
        # deploy converted config files
        p_dir.mkdir(exist_ok=True)
        try:
            if dynamic_config_file:
                dynamic_config_file = pathlib.Path(dynamic_config_file)
                auto_coding_setup = val_dynamic.setup_to_plane_config(p_dir, dynamic_config_file.parent, dynamic_config_file.name)
            else:
                auto_coding_setup = val_dynamic.setup_to_plane_config(p_dir)
        except Exception as exc:
            gt_gui.utils.push_popup(g, gt_gui.msg_box.msgbox, "Error converting setup", f"There was an error trying to load the dynamic validation\nsetup when creating the new validation plane:\n{exc}", gt_gui.msg_box.MsgBox.error, more=gt_gui.utils.get_traceback(type(exc), exc, exc.__traceback__))
            shutil.rmtree(p_dir)
            return
    # make plane
    p_def = plane.make_definition(p_type, name, p_dir, **extra)
    # store to file
    if not p_dir.is_dir():
        p_dir.mkdir()
    p_def.store_as_json(p_dir)
    # append to known planes
    g.study_config.planes.append(p_def)
    # deal with further setup for dynamic plane
    if is_dynamic:
        set_auto_coding_for_dynamic_validation_plane(g, p_def, auto_coding_setup)

def delete_plane(study_config: config.Study, plane: plane.Definition):
    # remove config directory for the plane
    path = config.guess_config_dir(study_config.working_directory)
    p_dir = path / plane.name
    if p_dir.is_dir():
        shutil.rmtree(p_dir)

def glasses_validator_plane_check_or_deploy_config(study_config: config.Study, pl: plane.Definition_GlassesValidator):
    if not isinstance(pl, plane.Definition_GlassesValidator) or pl.use_default:
        return
    # check if there are already are validation setup files
    working_dir = config.guess_config_dir(study_config.working_directory)/pl.name
    try:
        val_config.get_validation_setup(working_dir)
    except:
        # no config file, deploy
        default_poster.deploy_config(working_dir)
    else:
        # already exists, nothing to do
        pass

def glasses_validator_deploy_config(g, pl: plane.Definition_GlassesValidator):
    from . import gui
    g = typing.cast(gui.GUI,g)  # indicate type to typechecker
    if not isinstance(pl, plane.Definition_GlassesValidator):
        return
    working_dir = config.guess_config_dir(g.study_config.working_directory)/pl.name
    not_deployed = default_poster.deploy_config(working_dir)
    if not_deployed:
        buttons = {
            ifa6.ICON_FA_CHECK+" Yes": lambda: default_poster.deploy_config(working_dir,overwrite=True),
            ifa6.ICON_FA_CIRCLE_XMARK+" No": None
        }
        gt_gui.utils.push_popup(g, gt_gui.msg_box.msgbox, "Overwrite config?", "The following files already existed and were not deployed:\n- "+('\n- '.join(not_deployed))+"\nDo you want to overwrite these files?", gt_gui.msg_box.MsgBox.warn, buttons)

def make_recording_definition(study_config: config.Study, r_type: session.RecordingType, name: str):
    # append to defined recordings
    study_config.session_def.recordings.append(session.RecordingDefinition(name,r_type))
    # store config
    path = config.guess_config_dir(study_config.working_directory)
    study_config.session_def.store_as_json(path)

def delete_recording_definition(study_config: config.Study, recording: session.RecordingDefinition):
    # remove from defined recordings
    study_config.session_def.recordings = [r for r in study_config.session_def.recordings if r.name!=recording.name]
    # store config
    path = config.guess_config_dir(study_config.working_directory)
    study_config.session_def.store_as_json(path)

async def remove_recording_working_dir(project_dir: pathlib.Path, session: str, recording: str):
    rec_dir = project_dir / session / recording
    if rec_dir.is_dir():
        shutil.rmtree(rec_dir)

def make_individual_marker(study_config: config.Study, mark_id: int, aruco_dict_id: int, mark_detect_only: bool, mark_size: float=None):
    # append to defined recordings
    study_config.individual_markers.append(marker.Marker(mark_id, mark_detect_only, None if mark_detect_only else mark_size, aruco_dict_id))
    # store config
    study_config.store_as_json()

def delete_individual_marker(study_config: config.Study, mark: marker.Marker|None, index: int|None=None):
    # remove from defined individual markers
    if index is not None:
        del study_config.individual_markers[index]
    else:
        study_config.individual_markers = [m for m in study_config.individual_markers if m.id!=mark.id]
    # store config
    study_config.store_as_json()

def new_session_button(g, notify_func: typing.Callable[[str], None]|None = None):
    from . import gui
    g = typing.cast(gui.GUI,g)  # indicate type to typechecker
    new_sess_name = ''
    def _valid_sess_name():
        return new_sess_name and pathvalidate.is_valid_filename(new_sess_name, "auto") and g.sessions.get(new_sess_name, None) is None
    def _add_sess_popup():
        nonlocal new_sess_name
        imgui.dummy((30*imgui.calc_text_size('x').x,0))
        if imgui.begin_table("##new_sess_info",2):
            imgui.table_setup_column("##new_sess_infos_left", imgui.TableColumnFlags_.width_fixed)
            imgui.table_setup_column("##new_sess_infos_right", imgui.TableColumnFlags_.width_stretch)
            imgui.table_next_row()
            imgui.table_next_column()
            imgui.align_text_to_frame_padding()
            invalid = not _valid_sess_name()
            if invalid:
                imgui.push_style_color(imgui.Col_.text, colors.error)
            imgui.text("Session name")
            if invalid:
                imgui.pop_style_color()
            imgui.table_next_column()
            imgui.set_next_item_width(-1)
            _,new_sess_name = imgui.input_text("##new_sess_name",new_sess_name)
            imgui.end_table()

    def _make_session():
        make_session(g.project_dir, new_sess_name)
        if notify_func is not None:
            notify_func(new_sess_name)

    buttons = {
        ifa6.ICON_FA_CHECK+" Create session": (_make_session, lambda: not _valid_sess_name()),
        ifa6.ICON_FA_CIRCLE_XMARK+" Cancel": None
    }
    gt_gui.utils.push_popup(g, lambda: gt_gui.utils.popup("Add session", _add_sess_popup, buttons = buttons, button_keymap={0:imgui.Key.enter}, outside=False))

def make_session(project_dir: pathlib.Path, session_name: str):
    sess_dir = project_dir/session_name
    sess_dir.mkdir(exist_ok=True)
    session.get_action_states(sess_dir, for_recording=False, create_if_missing=True)

def open_url(path: str):
    # this works for files, folders and URLs
    if platform.os==platform.Os.Windows:
        os.startfile(path)
    else:
        if platform.os==platform.Os.Linux:
            open_util = "xdg-open"
        elif platform.os==platform.Os.MacOS:
            open_util = "open"
        async_thread.run(asyncio.create_subprocess_exec(
            open_util, path,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL
        ))

def open_folder(g, path: pathlib.Path):
    from . import gui
    g = typing.cast(gui.GUI,g)  # indicate type to typechecker
    if not path.is_dir():
        gt_gui.utils.push_popup(g, gt_gui.msg_box.msgbox, "Folder not found", f"The folder you're trying to open\n{path}\ncould not be found.", gt_gui.msg_box.MsgBox.warn)
        return
    open_url(str(path))

def remove_folder(folder: pathlib.Path):
    if folder.is_dir():
        shutil.rmtree(folder)

def show_action_options(g, session_name: str, rec_name: str|None, action: process.Action):
    from . import gui
    g = typing.cast(gui.GUI,g)  # indicate type to typechecker
    # NB: `show_visualization` is always an option, the others below are
    # shown when `show_visualization` is true (else they don't apply)
    match action:
        case process.Action.MAKE_GAZE_OVERLAY_VIDEO:
            options = {'show_visualization': [
                    False, 'Show visualization', 'Show a viewer that allows to follow the generation of the video. Each frame is shown with gaze overlaid as it is written into the video file.'
                ]}
        case process.Action.DETECT_MARKERS:
            options = {'show_visualization': [
                    False, 'Show visualization', 'Show a viewer that allows to follow the processing of the video. Each frame is shown overlaid with info about detected markers and planes.'
                ], 'visualization_show_rejected_markers': [
                    False, 'Show rejected markers', 'Rejected ArUco marker candidates are also shown in the viewer. Possibly useful for debug.'
                ]}
        case process.Action.GAZE_TO_PLANE:
            options = {'show_visualization': [
                    False, 'Show visualization', 'Show a viewer that visualizes the mapped gaze to plane. Each frame of the video is shown in the viewer, overlaid with info about detected planes and projected gaze.'
                ], 'show_planes': [
                    True, 'Show planes', "Additional viewer window(s) are opened, one per plane, with gaze in this plane's space is drawn on it."
                ], 'show_only_intervals': [
                    True, 'Show only intervals', 'Only the coded mapping episodes (if available) are shown in the viewer while the rest of the scene video is skipped past.'
                ]}
        case process.Action.MAKE_MAPPED_GAZE_VIDEO:
            options = {'show_visualization': [
                    False, 'Show visualization', 'The generated video(s) are shown in a viewer as they are created.'
                ]}
        case _:
            raise ValueError(f'Action option setting GUI not implemented for a {action} action')

    def set_action_options_popup():
        nonlocal options
        spacing = 2 * imgui.get_style().item_spacing.x
        color = (0.45, 0.09, 1.00, 1.00)
        imgui.push_font(g._icon_font)
        imgui.text_colored(color, ifa6.ICON_FA_CIRCLE_INFO)
        imgui.pop_font()
        imgui.same_line(spacing=spacing)

        imgui.begin_group()
        imgui.dummy((0,2*imgui.get_style().item_spacing.y))
        imgui.text_unformatted(f'Settings for this run of the {action.displayable_name} action\nSession: "{session_name}", recording: "{rec_name}":')
        imgui.dummy((0,1.5*imgui.get_style().item_spacing.y))
        _, options['show_visualization'][0] = imgui.checkbox(f'{options["show_visualization"][1]}##{session_name}_{rec_name}', options['show_visualization'][0])
        imgui.same_line()
        gt_gui.utils.draw_hover_text(options['show_visualization'][2])
        if options['show_visualization'][0]:
            for o in options:
                if o=='show_visualization':
                    continue
                _, options[o][0] = imgui.checkbox(f'{options[o][1]}##{session_name}_{rec_name}', options[o][0])
                imgui.same_line()
                gt_gui.utils.draw_hover_text(options[o][2])

        imgui.end_group()

    buttons = {
        ifa6.ICON_FA_CHECK+f" Continue##{session_name}_{rec_name}": lambda: g.launch_task(session_name, rec_name, action, **{o:options[o][0] for o in options}),
        ifa6.ICON_FA_CIRCLE_XMARK+f" Cancel##{session_name}_{rec_name}": None
    }

    # show configurable options for action
    gt_gui.utils.push_popup(g, lambda: gt_gui.utils.popup(f"Set options##{session_name}_{rec_name}", set_action_options_popup, buttons=buttons, button_keymap={0:imgui.Key.enter}, closable=True, outside=False))


def show_export_config(g, path: str|pathlib.Path, sessions: list[str]):
    from . import gui
    g = typing.cast(gui.GUI,g)  # indicate type to typechecker

    # for the sessions, see what we have to export
    to_export: dict[str,bool] = {}
    rec_dirs_val: list[pathlib.Path] = []
    rec_dirs_et_sync: list[pathlib.Path] = []
    for s_name in sessions:
        s = g.sessions.get(s_name,None)
        if s is None:
            continue
        if 'plane gaze' not in to_export:
            if annotation.Event.Trial in g.study_config.planes_per_episode and any((s.recordings[r].state[process.Action.GAZE_TO_PLANE]==process_pool.State.Completed for r in s.recordings)):
                to_export['plane gaze'] = True
        if recs:=[s.recordings[r].info.working_directory for r in s.recordings if s.recordings[r].state[process.Action.VALIDATE]==process_pool.State.Completed]:
            to_export['validation'] = True
            rec_dirs_val.extend(recs)
        if recs:=[s.recordings[r].info.working_directory for r in s.recordings if s.recordings[r].state[process.Action.SYNC_ET_TO_CAM]==process_pool.State.Completed]:
            to_export['eye tracker synchronization'] = True
            rec_dirs_et_sync.extend(recs)
        if 'gaze overlay video' not in to_export:
            if any((s.recordings[r].state[process.Action.MAKE_GAZE_OVERLAY_VIDEO]==process_pool.State.Completed for r in s.recordings)):
                to_export['gaze overlay video'] = True
        if 'mapped gaze video' not in to_export:
            if s.state[process.Action.MAKE_MAPPED_GAZE_VIDEO]==process_pool.State.Completed:
                to_export['mapped gaze video'] = True

    dq_df, dq_set = None, None
    if rec_dirs_val:
        dq_df, default_dq_type, dq_targets = export.collect_data_quality(rec_dirs_val, {p:f'{naming.validation_prefix}{p}_data_quality.tsv' for p in g.study_config.planes_per_episode[annotation.Event.Validate]}, col_for_parent='session')
        if dq_df is None:
            to_export.pop('validation', None)
        else:
            # prep config for validation export
            dq_set = {}

            # data quality type
            type_idx = dq_df.index.names.index('type')
            dq_set['dq_types'] = {k:False for k in sorted(list(dq_df.index.levels[type_idx]), key=lambda dq: dq.value)}
            for dq in DataQualityType:
                if g.study_config.validate_dq_types is not None and dq in g.study_config.validate_dq_types and dq in dq_set['dq_types']:
                    dq_set['dq_types'][dq] = True
            if not any(dq_set['dq_types'].values()):
                dq_set['dq_types'][default_dq_type] = True

            # targets
            dq_set['targets']     = {t:True for t in dq_targets}
            dq_set['targets_avg'] = False

            # other settings
            dq_set['include_data_loss'] = g.study_config.validate_include_data_loss

    def set_export_config_popup():
        nonlocal to_export, dq_set
        spacing = 2 * imgui.get_style().item_spacing.x
        color = (0.45, 0.09, 1.00, 1.00)
        imgui.push_font(g._icon_font)
        imgui.text_colored(color, ifa6.ICON_FA_CIRCLE_INFO)
        imgui.pop_font()
        imgui.same_line(spacing=spacing)

        imgui.begin_group()
        imgui.dummy((0,2*imgui.get_style().item_spacing.y))
        imgui.text_unformatted(f'Select below what you wish to export to the folder\n{path}')
        imgui.dummy((0,1.5*imgui.get_style().item_spacing.y))
        for e in to_export:
            _, to_export[e] = imgui.checkbox(e, to_export[e])

        if 'validation' in to_export and to_export['validation'] and imgui.tree_node('Validation export settings'):
            right_width = hello_imgui.dpi_window_size_factor()*90
            if len(dq_set['dq_types'])>1:
                if imgui.tree_node('Data quality types'):
                    imgui.text_unformatted("Indicate which type(s) of\ndata quality to export.")
                    if imgui.begin_table("##export_popup_validation", columns=2, flags=imgui.TableFlags_.no_clip):
                        imgui.table_setup_column("##settings_validation_left", imgui.TableColumnFlags_.width_stretch)
                        imgui.table_setup_column("##settings_validation_right", imgui.TableColumnFlags_.width_fixed)
                        imgui.table_next_row()
                        imgui.table_set_column_index(1)  # Right
                        imgui.dummy((right_width, 1))

                        for dq in dq_set['dq_types']:
                            imgui.table_next_row()
                            imgui.table_next_column()
                            imgui.align_text_to_frame_padding()
                            t,ht = get_DataQualityType_explanation(dq)
                            imgui.text(t)
                            gt_gui.utils.draw_hover_text(ht, text="")
                            imgui.table_next_column()
                            _, dq_set['dq_types'][dq] = imgui.checkbox(f"##{dq.name}", dq_set['dq_types'][dq])

                        imgui.end_table()
                        imgui.spacing()
                    imgui.tree_pop()

            if imgui.tree_node('Targets'):
                imgui.text_unformatted("Indicate for which target(s) you\nwant to export data quality metrics.")
                if imgui.begin_table("##export_popup_targets", columns=2, flags=imgui.TableFlags_.no_clip):
                    imgui.table_setup_column("##settings_targets_left", imgui.TableColumnFlags_.width_stretch)
                    imgui.table_setup_column("##settings_targets_right", imgui.TableColumnFlags_.width_fixed)
                    imgui.table_next_row()
                    imgui.table_set_column_index(1)  # Right
                    imgui.dummy((right_width, 1))

                    for t in dq_set['targets']:
                        imgui.table_next_row()
                        imgui.table_next_column()
                        imgui.align_text_to_frame_padding()
                        imgui.text(f"target {t}:")
                        imgui.table_next_column()
                        _, dq_set['targets'][t] = imgui.checkbox(f"##target_{t}", dq_set['targets'][t])

                    imgui.end_table()
                    imgui.spacing()
                imgui.tree_pop()

            if imgui.begin_table("##export_popup_targets_avg", columns=2, flags=imgui.TableFlags_.no_clip):
                imgui.table_setup_column("##settings_targets_avg_left", imgui.TableColumnFlags_.width_stretch)
                imgui.table_setup_column("##settings_targets_avg_right", imgui.TableColumnFlags_.width_fixed)
                imgui.table_next_row()
                imgui.table_set_column_index(1)  # Right
                imgui.dummy((right_width, 1))

                imgui.table_next_row()
                imgui.table_next_column()
                imgui.align_text_to_frame_padding()
                imgui.text("Average over selected targets:")
                imgui.table_next_column()
                _, dq_set['targets_avg'] = imgui.checkbox("##average_over_targets", dq_set['targets_avg'])

                imgui.end_table()
                imgui.spacing()

            imgui.tree_pop()

        imgui.end_group()

    def launch_export():
        exp = []
        if 'plane gaze' in to_export and to_export['plane gaze']:
            exp.append('plane_gaze')
        if 'gaze overlay video' in to_export and to_export['gaze overlay video']:
            exp.append('gaze_overlay_video')
        if 'mapped gaze video' in to_export and to_export['mapped gaze video']:
            exp.append('mapped_gaze_video')
        for s in sessions:
            g.launch_task(s, None, process.Action.EXPORT_TRIALS, export_path=path, to_export=exp)

        # export data quality for all recordings in all sessions
        if 'validation' in to_export and to_export['validation']:
            dq_types = [dq for dq in dq_set['dq_types'] if dq_set['dq_types'][dq]]
            targets  = [t for t in dq_set['targets'] if dq_set['targets'][t]]
            export.summarize_and_store_data_quality(dq_df, path/'data_quality.tsv', dq_types, targets, dq_set['targets_avg'], dq_set['include_data_loss'])

        # export et sync for all recordings in all sessions
        if 'eye tracker synchronization' in to_export and to_export['eye tracker synchronization']:
            export.export_et_sync(rec_dirs_et_sync, naming.VOR_sync_file, path/'et_sync.tsv')

    buttons = {
        ifa6.ICON_FA_CHECK+f" Continue": (launch_export, lambda: not any((to_export[e] for e in to_export))),
        ifa6.ICON_FA_CIRCLE_XMARK+f" Cancel": None
    }

    # ask what to export
    gt_gui.utils.push_popup(g, lambda: gt_gui.utils.popup(f"Set what to export", set_export_config_popup, buttons=buttons, button_keymap={0:imgui.Key.enter}, closable=True, outside=False))


async def _show_addable_recordings(g, rec_getter: typing.Callable[[],list[recording.Recording|camera_recording.Recording]], dev_type: session.RecordingType, dev: eyetracker.EyeTracker|None, sessions: list[str], generic_device_name: str=None):
    from . import gui
    g = typing.cast(gui.GUI,g)  # indicate type to typechecker

    if dev_type==session.RecordingType.Camera:
        dev_rec_lbl = 'Camera recordings'
    else:
        if dev==eyetracker.EyeTracker.Generic:
            dev_rec_lbl = f'generic recordings for a {generic_device_name} eye tracker'
        else:
            dev_rec_lbl = f'{dev.value} eye tracker recordings'

    # notify we're preparing the recordings to be opened
    def prepping_recs_popup():
        spacing = 2 * imgui.get_style().item_spacing.x
        color = (0.45, 0.09, 1.00, 1.00)
        imgui.push_font(g._icon_font)
        imgui.text_colored(color, ifa6.ICON_FA_CIRCLE_INFO)
        imgui.pop_font()
        imgui.same_line(spacing=spacing)

        imgui.begin_group()
        imgui.dummy((0,2*imgui.get_style().item_spacing.y))
        text = f'Searching the path(s) you provided for {dev_rec_lbl}.'
        imgui.text_unformatted(text)
        imgui.dummy((0,3*imgui.get_style().item_spacing.y))
        text_size = imgui.calc_text_size(text)
        size_mult = hello_imgui.dpi_window_size_factor()
        spinner_radii = [x*size_mult for x in [22, 16, 10]]
        imgui.set_cursor_pos_x(imgui.get_cursor_pos_x()+(text_size.x-2*spinner_radii[0])/2)
        imspinner.spinner_ang_triple('waitSpinner', *spinner_radii, 3.5*size_mult, c1=imgui.get_style_color_vec4(imgui.Col_.text), c2=colors.warning, c3=imgui.get_style_color_vec4(imgui.Col_.text))
        imgui.dummy((0,2*imgui.get_style().item_spacing.y))
        imgui.end_group()

    gt_gui.utils.push_popup(g, lambda: gt_gui.utils.popup("Preparing import", prepping_recs_popup, buttons=None, button_keymap={0:imgui.Key.enter}, closable=False, outside=False))

    # step 1, find what recordings of this type of eye tracker/camera files are in the path
    recs = rec_getter()
    all_recs: list[recording.Recording] = []
    dup_recs: list[recording.Recording] = []
    known_rec_dirs = [g.sessions[s].recordings[r].info.get_source_directory() for s in g.sessions for r in g.sessions[s].recordings if g.sessions[s].recordings[r].definition.type==dev_type]
    known_rec_names= [g.sessions[s].recordings[r].info.name for s in g.sessions for r in g.sessions[s].recordings if g.sessions[s].recordings[r].definition.type==dev_type]
    for rec in recs:
        # skip duplicates. For eye tracker recordings, this checks for the source folder. For camera recordings, its checks for the full file path
        if dev_type==session.RecordingType.Camera:
            dup = rec.get_source_directory()/rec.video_file in (g.sessions[s].recordings[r].info.get_source_directory()/g.sessions[s].recordings[r].info.video_file for s in g.sessions for r in g.sessions[s].recordings if g.sessions[s].recordings[r].definition.type==dev_type)
        elif dev_type==session.RecordingType.Eye_Tracker:
            dup = any([a and b for a,b in zip([rec.get_source_directory()==d for d in known_rec_dirs],[rec.name==n for n in known_rec_names])])
        if dup:
            dup_recs.append(rec)
        else:
            all_recs.append(rec)

    # get ready to show result
    # 1. remove prepping recordings popup
    del g.popup_stack[-1]

    # 2. if nothing importable found, notify
    if not all_recs:
        if dup_recs:
            msg = f"{dev_rec_lbl} were found in the specified import paths, but could not be imported as they are already part of sessions in this gazeMapper project."
            more= "Duplicates that were not imported:\n"+('\n'.join([str(r.get_source_directory()) for r in dup_recs]))
        else:
            msg = f"No {dev_rec_lbl} were found among the specified import paths."
            more = None

        gt_gui.utils.push_popup(g, gt_gui.msg_box.msgbox, "Nothing to import", msg, gt_gui.msg_box.MsgBox.warn, more=more)
        return

    # 3. if something importable found, show to user so they can assign them to recordings in sessions
    recordings_to_add = {i:r for i,r in enumerate(all_recs)}
    recording_assignment: dict[str, dict[str, int]] = {}
    rec_names = [r.name for r in g.study_config.session_def.recordings]
    selected_slot: tuple[str,str] = None

    def _recording_context_menu(iid: int) -> bool:
        nonlocal selected_slot
        if imgui.selectable(ifa6.ICON_FA_FOLDER_OPEN + f" Open folder##{iid}", False)[0]:
            open_folder(g, recordings_to_add[iid].get_source_directory())
        if selected_slot is not None and imgui.selectable(ifa6.ICON_FA_ARROW_LEFT + f" Assign to selected recording", False)[0]:
            recording_assignment[selected_slot[0]][selected_slot[1]] = iid
            selected_slot = None
            recording_list.require_sort = True
        return False

    def _add_new_session(new_sess: str):
        sessions.append(new_sess)

    recording_lock = threading.Lock()
    recording_list = gt_gui.recording_table.RecordingTable(recordings_to_add, recording_lock, None, item_context_callback=_recording_context_menu)
    recording_list.set_local_item_remover()
    recording_list.set_act_as_drag_drop_source(True)
    if dev_type==session.RecordingType.Camera:
        recording_list.show_hide_columns({'Eye Tracker': False, 'Participant': False, 'Source Directory': True, 'Video File': True})
    not_assigned_filter = gt_gui.recording_table.Filter(lambda iid, _: iid not in (recording_assignment[s][r] for s in recording_assignment for r in recording_assignment[s]))
    recording_list.add_filter(not_assigned_filter)
    def list_recs_popup():
        nonlocal selected_slot
        spacing = 2 * imgui.get_style().item_spacing.x
        imgui.same_line(spacing=spacing)

        imgui.text_unformatted("Select which recordings you would like to import. Assign a recording from the list on the right to a recording in a session on the left by dragging and dropping it. Add new sessions if needed")
        imgui.dummy((0,1*imgui.get_style().item_spacing.y))

        size_mult = hello_imgui.dpi_window_size_factor()
        imgui.begin_child("##main_frame_adder", size=(1260*size_mult,min(700*size_mult,max(400*size_mult,(len(recording_list.recordings)+2)*imgui.get_frame_height_with_spacing()))))
        imgui.begin_child("##session_list", size=(600*size_mult,0))
        if imgui.button('+ new session'):
            new_session_button(g, _add_new_session)
        for s in sessions:
            sess = g.sessions.get(s, None)
            if sess is None:
                continue
            if sess.name not in recording_assignment:
                recording_assignment[sess.name] = {}
            if imgui.tree_node_ex(sess.name):
                table_opened = False
                for r in rec_names:
                    if sess.definition.get_recording_def(r).type!=dev_type:
                        continue
                    disable = False
                    rec: recording.Recording = None
                    if r in sess.recordings:
                        rec = sess.recordings[r].info
                        disable = True
                    elif r in recording_assignment[sess.name]:
                        rec = recordings_to_add[recording_assignment[sess.name][r]]
                    has_rec = rec is not None
                    if disable:
                        imgui.begin_disabled()
                    if table_opened or imgui.begin_table(f'##{sess.name}', 4, imgui.TableFlags_.sizing_fixed_fit):
                        table_opened = True
                        imgui.table_next_column()
                        imgui.align_text_to_frame_padding()
                        if not disable:
                            selected = False if not selected_slot else (sess.name,r)==selected_slot
                            interacted, was_selected = imgui.selectable(r, selected, imgui.SelectableFlags_.span_all_columns|imgui.SelectableFlags_.allow_overlap)
                            if has_rec:
                                if imgui.begin_popup_context_item(f"##{sess.name}_{r}_context"):
                                    if imgui.selectable(ifa6.ICON_FA_ARROW_RIGHT + f" Unassign this recording", False)[0]:
                                        recording_assignment[sess.name].pop(r, None)
                                        recording_list.require_sort = True
                                    imgui.end_popup()
                            else:
                                if interacted:
                                    if was_selected:
                                        selected_slot = (sess.name,r)
                                    else:
                                        selected_slot = None
                                if imgui.begin_drag_drop_target():
                                    payload = imgui.accept_drag_drop_payload_py_id("RECORDING")
                                    if payload is not None:
                                        recording_assignment[sess.name][r] = payload.data_id
                                        recording_list.require_sort = True
                                    imgui.end_drag_drop_target()
                        else:
                            imgui.text(r)
                        imgui.table_next_column()
                        if has_rec:
                            if dev_type==session.RecordingType.Eye_Tracker:
                                recording_list.draw_eye_tracker_widget(rec, align=True)
                            else:
                                imgui.text(rec.video_file)
                        imgui.table_next_column()
                        if has_rec:
                            if dev_type==session.RecordingType.Eye_Tracker:
                                imgui.text(rec.name)
                            else:
                                imgui.text(str(rec.source_directory))
                        else:
                            imgui.text('drop recording here to assign')
                        imgui.table_next_column()
                        if has_rec and dev_type==session.RecordingType.Eye_Tracker:
                            imgui.text(rec.participant)
                    if disable:
                        imgui.end_disabled()

                if table_opened:
                    table_opened = False
                    imgui.end_table()
                imgui.tree_pop()

        imgui.end_child()
        imgui.same_line()
        imgui.begin_child("##import_source")
        imgui.begin_child("##recording_list_frame", size=(0,-imgui.get_frame_height_with_spacing()), window_flags=imgui.WindowFlags_.horizontal_scrollbar)
        recording_list.draw()
        imgui.end_child()
        imgui.begin_child("##recording_list_bottombar_frame")
        recording_list.filter_box_text, recording_list.require_sort = \
            draw_filterbar(g, recording_list.filter_box_text, recording_list.require_sort)
        imgui.end_child()
        imgui.end_child()
        imgui.end_child()

    buttons = {
        ifa6.ICON_FA_CHECK+" Continue": lambda: async_thread.run(_import_recordings(g, recordings_to_add, recording_assignment, generic_device_name)),
        ifa6.ICON_FA_CIRCLE_XMARK+" Cancel": None
    }
    gt_gui.utils.push_popup(g, lambda: gt_gui.utils.popup("Assign and import recordings", list_recs_popup, buttons=buttons, button_keymap={0:imgui.Key.enter}, closable=True, outside=False))

async def _import_recordings(g, recordings: list[recording.Recording|camera_recording.Recording], recording_assignment: dict[str, dict[str, int]], generic_device_name: str|None):
    from . import gui
    g = typing.cast(gui.GUI,g)  # indicate type to typechecker
    for s in recording_assignment:
        sess = g.sessions.get(s, None)
        if sess is None:
            continue
        for r in recording_assignment[s]:
            rec = recordings[recording_assignment[s][r]]
            # first create recording only in memory
            sess.add_recording_from_info(r, rec)
            # then launch import task
            g.launch_task(s, r, process.Action.IMPORT, generic_device_name=generic_device_name)

def draw_filterbar(g, filter_box_text: str, require_sort: bool):
    from . import gui
    g = typing.cast(gui.GUI,g)  # indicate type to typechecker
    imgui.set_next_item_width(-imgui.FLT_MIN)
    _, value = imgui.input_text_with_hint(f"##filterbar", "Start typing to filter the list", filter_box_text, flags=imgui.InputTextFlags_.enter_returns_true)
    if imgui.begin_popup_context_item(f"##filterbar_context"):
        # Right click = more options context menu
        if imgui.selectable(ifa6.ICON_FA_CLIPBOARD+" Paste", False)[0]:
            value += imgui.get_clipboard_text() or ""
        imgui.separator()
        if imgui.selectable(ifa6.ICON_FA_CIRCLE_INFO+" More info", False)[0]:
            gt_gui.utils.push_popup(g,
                gt_gui.msg_box.msgbox, "About the filter bar",
                "This is the filter bar. By typing inside it you can search your recording list inside the eye tracker, name, participant and project properties.",
                gt_gui.msg_box.MsgBox.info
            )
        imgui.end_popup()
    if value != filter_box_text:
        filter_box_text = value
        require_sort = True

    return filter_box_text, require_sort

def add_eyetracking_recordings(g, paths: list[pathlib.Path], sessions: list[str]):
    from . import gui
    g = typing.cast(gui.GUI,g)  # indicate type to typechecker
    combo_value = 0
    invalid = False
    generic_et_idx = -1
    generic_et_name = ''
    eye_tracker = eyetracker.EyeTracker(eyetracker.eye_tracker_names[combo_value])
    sessions = get_and_filter_eligible_sessions(g, sessions, session.RecordingType.Eye_Tracker)

    def add_recs_popup():
        nonlocal combo_value, eye_tracker, generic_et_idx, generic_et_name
        spacing = 2 * imgui.get_style().item_spacing.x
        color = (0.45, 0.09, 1.00, 1.00)
        imgui.push_font(g._icon_font)
        imgui.text_colored(color, ifa6.ICON_FA_CIRCLE_INFO)
        imgui.pop_font()
        imgui.same_line(spacing=spacing)

        imgui.begin_group()
        imgui.dummy((0,2*imgui.get_style().item_spacing.y))
        imgui.text_unformatted("For which eye tracker would you like to import recordings?")
        imgui.dummy((0,3*imgui.get_style().item_spacing.y))
        full_width = imgui.get_content_region_avail().x
        imgui.push_item_width(full_width*.4)
        imgui.set_cursor_pos_x(full_width*.3)
        changed, combo_value = imgui.combo("##select_eye_tracker", combo_value, eyetracker.eye_tracker_names)
        if changed:
            eye_tracker = eyetracker.EyeTracker(eyetracker.eye_tracker_names[combo_value])
            if eye_tracker!=eyetracker.EyeTracker.Generic:
                invalid = False
                generic_et_idx = -1
                generic_et_name = ''
        imgui.pop_item_width()
        imgui.dummy((0,2*imgui.get_style().item_spacing.y))
        if eye_tracker==eyetracker.EyeTracker.Generic:
            imgui.text_unformatted("Choose which generic eye tracker you want")
            imgui.text_unformatted("to import data for:")
            imgui.dummy((0,1.5*imgui.get_style().item_spacing.y))
            invalid = not g.study_config.import_known_custom_eye_trackers
            if invalid:
                imgui.push_style_color(imgui.Col_.text, colors.error)
                imgui.text("No custom eye trackers configured, fix project settings")
                imgui.pop_style_color()
            else:
                full_width = imgui.get_content_region_avail().x
                imgui.set_cursor_pos_x(full_width*.3)
                imgui.push_item_width(full_width*.6)
                changed, generic_et_idx = imgui.combo('##generic_et_name_selector', generic_et_idx, g.study_config.import_known_custom_eye_trackers)
                if changed:
                    generic_et_name = g.study_config.import_known_custom_eye_trackers[generic_et_idx]
                imgui.pop_item_width()
            imgui.dummy((0,2*imgui.get_style().item_spacing.y))

        imgui.end_group()

        return combo_value, eye_tracker

    buttons = {
        ifa6.ICON_FA_CHECK+" Continue": (lambda: async_thread.run(_show_addable_recordings(g, lambda: recording.find_recordings(paths, eye_tracker, None if not generic_et_name else generic_et_name), session.RecordingType.Eye_Tracker, eye_tracker, sessions, None if not generic_et_name else generic_et_name)), lambda: invalid or (eye_tracker==eyetracker.EyeTracker.Generic and generic_et_idx==-1)),
        ifa6.ICON_FA_CIRCLE_XMARK+" Cancel": None
    }

    # ask what type of eye tracker we should be looking for
    gt_gui.utils.push_popup(g, lambda: gt_gui.utils.popup("Select eye tracker", add_recs_popup, buttons=buttons, button_keymap={0:imgui.Key.enter}, closable=True, outside=False))

def camera_show_glob_filter_config(g, paths, sessions):
    from . import gui
    g = typing.cast(gui.GUI,g)  # indicate type to typechecker
    glob_filter = '*.mp4,*.mov,*.avi,*.mkv'

    def setting_popup():
        nonlocal glob_filter
        spacing = 2 * imgui.get_style().item_spacing.x
        color = (0.45, 0.09, 1.00, 1.00)
        imgui.push_font(g._icon_font)
        imgui.text_colored(color, ifa6.ICON_FA_CIRCLE_INFO)
        imgui.pop_font()
        imgui.same_line(spacing=spacing)

        imgui.begin_group()
        imgui.dummy((0,2*imgui.get_style().item_spacing.y))
        glob_filter = _set_glob_filter_for_camera(glob_filter)
        imgui.end_group()

    buttons = {
        ifa6.ICON_FA_CHECK+" Continue": lambda: add_camera_recordings(g, paths, glob_filter, sessions),
        ifa6.ICON_FA_CIRCLE_XMARK+" Cancel": None
    }
    gt_gui.utils.push_popup(g, lambda: gt_gui.utils.popup("Select device", setting_popup, buttons=buttons, button_keymap={0:imgui.Key.enter}, closable=True, outside=False))

def add_camera_recordings(g, paths: list[pathlib.Path], glob_filter: str, sessions: list[str]):
    from . import gui
    g = typing.cast(gui.GUI,g)  # indicate type to typechecker
    sessions = get_and_filter_eligible_sessions(g, sessions, session.RecordingType.Camera)
    async_thread.run(_show_addable_recordings(g, lambda: _find_camera_recordings(paths, glob_filter), session.RecordingType.Camera, None, sessions))

def _find_camera_recordings(paths: list[pathlib.Path], glob_filter: str) -> list[camera_recording.Recording]:
    extensions = {'.'+x.strip('.* ') for x in glob_filter.split(',')}
    video_paths: list[pathlib.Path] = []
    for p in paths:
        if p.is_file():
            if p.suffix in extensions:
                video_paths.append(p)
        elif p.is_dir():
            video_paths.extend([pth.resolve() for pth in p.glob("**/*") if pth.suffix in extensions])
    # turn into list of recordings
    return [camera_recording.Recording('',p.name,p.parent, duration=video_utils.get_video_duration(p)) for p in video_paths]

def get_and_filter_eligible_sessions(g, sessions: list[str], dev_type:session.RecordingType) -> list[str]:
    from . import gui
    g = typing.cast(gui.GUI,g)  # indicate type to typechecker
    if not sessions:
        sessions: list[str] = []
        for s in g.sessions:
            if not g.sessions[s].missing_recordings(dev_type):
                continue
            sessions.append(s)
    else:
        sessions = [s for s in sessions if g.sessions[s].missing_recordings(dev_type)]
    return sessions

def _set_glob_filter_for_camera(glob_filter):
    imgui.text_unformatted("Extension filter to search for video files in")
    imgui.text_unformatted("the selected paths (leave empty to not filter)")
    imgui.dummy((0,1.5*imgui.get_style().item_spacing.y))
    full_width = imgui.get_content_region_avail().x
    imgui.push_item_width(full_width*.8)
    _, glob_filter = imgui.input_text('##file_filter', glob_filter)
    imgui.pop_item_width()
    imgui.dummy((0,2*imgui.get_style().item_spacing.y))
    return glob_filter

def add_recordings(g, paths: list[pathlib.Path], sessions: list[str]):
    from . import gui
    g = typing.cast(gui.GUI,g)  # indicate type to typechecker
    options: list[session.RecordingType] = []
    if any((r.type==session.RecordingType.Eye_Tracker for r in g.study_config.session_def.recordings)):
        options.append(session.RecordingType.Eye_Tracker)
    if any((r.type==session.RecordingType.Camera for r in g.study_config.session_def.recordings)):
        options.append(session.RecordingType.Camera)
    if not options:
        return
    combo_value = 0
    dev_type = options[combo_value]
    glob_filter = '*.mp4,*.mov,*.avi,*.mkv'

    def choose_dev_popup():
        nonlocal combo_value, dev_type, glob_filter
        spacing = 2 * imgui.get_style().item_spacing.x
        color = (0.45, 0.09, 1.00, 1.00)
        imgui.push_font(g._icon_font)
        imgui.text_colored(color, ifa6.ICON_FA_CIRCLE_INFO)
        imgui.pop_font()
        imgui.same_line(spacing=spacing)

        imgui.begin_group()
        imgui.dummy((0,2*imgui.get_style().item_spacing.y))
        imgui.text_unformatted("For which device would you like to import recordings?")
        imgui.dummy((0,1.5*imgui.get_style().item_spacing.y))
        full_width = imgui.get_content_region_avail().x
        imgui.push_item_width(full_width*.4)
        imgui.set_cursor_pos_x(full_width*.3)
        changed, combo_value = imgui.combo("##select_device", combo_value, [d.value for d in options])
        if changed:
            dev_type = options[combo_value]
        imgui.pop_item_width()
        imgui.dummy((0,2*imgui.get_style().item_spacing.y))
        if dev_type==session.RecordingType.Camera:
            glob_filter = _set_glob_filter_for_camera(glob_filter)

        imgui.end_group()

        return combo_value, dev_type

    def _run():
        match dev_type:
            case session.RecordingType.Eye_Tracker:
                add_eyetracking_recordings(g, paths, sessions)
            case session.RecordingType.Camera:
                add_camera_recordings(g, paths, glob_filter, sessions)

    if len(options)==1 and options[0]==session.RecordingType.Eye_Tracker:
        # no need to show selection popup, as there is only one choice and nothing to configure for that choice
        _run()
        return

    buttons = {
        ifa6.ICON_FA_CHECK+" Continue": _run,
        ifa6.ICON_FA_CIRCLE_XMARK+" Cancel": None
    }

    # ask what type of device recordings we want to import
    gt_gui.utils.push_popup(g, lambda: gt_gui.utils.popup("Select device", choose_dev_popup, buttons=buttons, button_keymap={0:imgui.Key.enter}, closable=True, outside=False))



import typing
import typeguard
import inspect

from glassesTools import aruco, json, marker as gt_marker

from . import naming, type_utils


class Marker:
    @typeguard.typechecked
    def __init__(self,
                 m_id               : int,
                 detect_only        : bool,         # if true, pose will not be determined and only marker presence is detected. That means marker size is not needed
                 size               : float|None                = None,
                 aruco_dict_id      : type_utils.ArucoDictType  = aruco.default_dict,
                 marker_border_bits : int                       = 1
                 ):
        self.id                 = m_id
        self.detect_only        = detect_only
        self.size               = size
        self.aruco_dict_id      = aruco_dict_id
        self.marker_border_bits = marker_border_bits

    def _to_dict(self) -> dict[str,typing.Any]:
        # N.B.: print dictionary names for markers instead of hard to understand integer id
        # always store dictionary even if its the default, so its easier to read the config file by eye
        out = {'id': self.id, 'aruco_dict': aruco.dict_id_to_str[self.aruco_dict_id], 'detect_only': self.detect_only}
        for f in ['size', 'marker_border_bits']:
            if (val:=getattr(self,f))!=marker_defaults[f]:
                out[f] = val
        return out

    @staticmethod
    def _from_dict(kwargs: dict[str,typing.Any]):
        # backwards compatibility
        if 'detect_only' not in kwargs:
            kwargs['detect_only'] = False
        if 'id' in kwargs:
            kwargs['m_id'] = kwargs.pop('id')
        if 'aruco_dict' in kwargs:
            kwargs['aruco_dict_id'] = kwargs.pop('aruco_dict')
        if 'aruco_dict_id' in kwargs:
            # dictionary names might be stored as strings, turn back into int
            if isinstance(kwargs['aruco_dict_id'],str):
                kwargs['aruco_dict_id'] = aruco.str_to_dict_id(kwargs['aruco_dict_id'])
        return Marker(**kwargs)

json.register_type(json.TypeEntry(Marker,'__marker.Marker__', Marker._to_dict, Marker._from_dict))
# get defaults for default argument of Marker constructor
_params = inspect.signature(Marker.__init__).parameters
marker_defaults = {k:d for k in _params if (d:=_params[k].default)!=inspect._empty}
marker_parameter_types = {k:_params[k].annotation for k in _params if k!='self'}
del _params

def get_marker_setup(marker: Marker) -> aruco.MarkerSetup:
    return aruco.MarkerSetup(aruco_detector_params = {
                                    'markerBorderBits': marker.marker_border_bits
                                },
                                detect_only = marker.detect_only,
                                size= marker.size
                            )

def get_setup_for_markers(markers: list[Marker]) -> dict[gt_marker.MarkerID,aruco.MarkerSetup]:
    out = {}
    for m in markers:
        out[gt_marker.MarkerID(m.id, m.aruco_dict_id)] = get_marker_setup(m)
    return out


from watchfiles import Change, DefaultFilter, awatch
import asyncio
import pathlib
from typing import Callable

from glassesTools import utils

class ChangeFilter(DefaultFilter):
    def __init__(self, extensions={'.json', '.gazeMapper'}, do_report_directories:bool=False, exclude_paths:set[pathlib.Path]=None, do_report_files=False):
        super().__init__(ignore_paths=exclude_paths)
        self.extensions             = extensions
        self.do_report_directories  = do_report_directories
        self.exclude_paths          = exclude_paths or set()
        self.do_report_files        = do_report_files

        # to be able to correctly detect directories, we have a problem. We cannot use is_dir() on a path, would always be false for deletes
        # so instead keep a list of all known directories. For adds and modifies go out to the file system, for deletes, check this list
        # checking for a directory with PurePath.suffix is not reliable as it would filter out all directories with a dot (.) in their name
        self.base_dir: pathlib.Path = None
        self._known_dirs: set[pathlib.Path] = None

    def set_base_dir(self, base_dir: pathlib.Path):
        self.base_dir = base_dir
        self._known_dirs = set(utils.fast_scandir(self.base_dir))
        # filter out excluded paths
        for p in self.exclude_paths:
            self._known_dirs = {kp for kp in self._known_dirs if not kp.is_relative_to(p)}

    def __call__(self, change: Change, path: str) -> bool:
        if not super().__call__(change, path):
            return False

        # handle deletes separately from modifies and adds
        path = pathlib.Path(path)
        if change==Change.deleted:
            if path in self._known_dirs:
                self._known_dirs.discard(path)
                return self.do_report_directories
            elif not self.do_report_files:
                return False
            else:
                return path.suffix in self.extensions
        else:
            if path.is_dir():
                self._known_dirs.add(path)
                return self.do_report_directories
            elif not self.do_report_files:
                return False
            else:
                return path.suffix in self.extensions

async def watch_and_report_changes(path: pathlib.Path, callback: Callable, stop_event: asyncio.Event, watch_filter=ChangeFilter()):
    watch_filter.set_base_dir(path)

    async for changes in awatch(path, debounce=500, watch_filter=watch_filter, stop_event=stop_event):
        for change_type,change_path in changes:
            callback(pathlib.Path(change_path), change_type.raw_str())